"""
Branch-and-bound dispatcher implementation.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import array
import time
import array
import collections
import os
import socket
import logging
import marshal
import math

from pybnb.configuration import config
from pybnb.common import (minimize,
                          maximize,
                          inf,
                          TerminationCondition,
                          _termination_condition_to_int)
from pybnb.misc import get_gap_labels
from pybnb.dispatcher_proxy import (ProcessType,
                                    DispatcherAction,
                                    DispatcherResponse,
                                    DispatcherProxy)
from pybnb.node import _SerializedNode
from pybnb.problem import _SolveInfo
from pybnb.mpi_utils import Message
from pybnb.priority_queue import PriorityQueueFactory

try:
    import mpi4py
except ImportError:                               #pragma:nocover
    pass

from sortedcontainers import SortedList

import six

class DispatcherQueueData(object):
    """A namedtuple storing data that can be used to
    initialize a dispatcher queue.

    Attributes
    ----------
    nodes : list
        A list of :class:`Node <pybnb.node.Node>` objects.
    worst_terminal_bound : float or None
        The worst bound of any node where branching did not
        continue.
    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem that produced
        this queue.
    """
    __slots__ = ("nodes",
                 "worst_terminal_bound",
                 "sense")
    def __init__(self,
                 nodes,
                 worst_terminal_bound,
                 sense):
        self.nodes = nodes
        self.worst_terminal_bound = worst_terminal_bound
        self.sense = sense
        assert sense in (minimize, maximize)

    def bound(self):
        """Returns the global bound defined by this queue data."""
        bound = None
        if len(self.nodes) > 0:
            if self.sense == minimize:
                bound = min(n_.bound for n_ in self.nodes)
            else:
                assert self.sense == maximize
                bound = max(n_.bound for n_ in self.nodes)
        if self.worst_terminal_bound is not None:
            if bound is not None:
                if self.sense == minimize:
                    bound = min(bound,
                                self.worst_terminal_bound)
                else:
                    assert self.sense == maximize
                    bound = max(bound,
                                self.worst_terminal_bound)
            else:
                bound = self.worst_terminal_bound
        return bound

class StatusPrinter(object):
    """Logs status information about the branch-and-bound
    solve.

    Parameters
    ----------
    dispatcher : :class:`pybnb.dispatcher.Dispatcher`
        The central dispatcher that will be monitored.
    log : :class:`logging.Logger`
        A log object where solver output should be sent.
    log_interval_seconds : float
        The approximate maximum time (in seconds) between
        solver log updates. More time may pass between log
        updates if no updates have been received from any
        workers, and less time may pass if a new incumbent
        is found. (default: 1.0)
    """

    def __init__(self,
                 dispatcher,
                 log,
                 log_interval_seconds=1.0):
        assert log_interval_seconds >= 0
        self._dispatcher = dispatcher
        self._log_interval_seconds = log_interval_seconds
        self._log = log

        percent_relative_gap_tol = 1e-6
        if (self._dispatcher.converger.relative_gap is not None) and \
           self._dispatcher.converger.relative_gap != 0:
            percent_relative_gap_tol = 100.0 * \
                self._dispatcher.converger.relative_gap
        rgap_str_length, rgap_label_str, rgap_number_str = \
            get_gap_labels(percent_relative_gap_tol, key="rgap")

        absolute_gap_tol = 1e-8
        if (self._dispatcher.converger.absolute_gap is not None) and \
           self._dispatcher.converger.absolute_gap != 0:
            absolute_gap_tol = \
                self._dispatcher.converger.absolute_gap
        agap_str_length, agap_label_str, agap_number_str = \
            get_gap_labels(absolute_gap_tol, key="agap", format='g')

        assert rgap_str_length >= 10
        assert agap_str_length >= 10
        extra_space = (rgap_str_length-10) + (agap_str_length-10)
        extra_space_left = extra_space // 2
        extra_space_right = (extra_space // 2) + (extra_space % 2)
        self._lines = ("--------------------"
                       "--------------------"
                       "--------------------"
                       "--------------------"
                       "--------------------"
                       "----------------")+("-"*extra_space)
        self._initial_header_line = \
            (self._lines + "\n"
             "         Nodes        |" + \
             (" "*extra_space_left) + \
             "                   Objective Bounds                    " + \
             (" "*extra_space_right) + \
             "|              Work              ")
        self._header_line = \
            (" {explored:>9} {unexplored:>9}  |{objective:>15} "
             "{bound:>15} "+rgap_label_str+"  "+agap_label_str+" |{runtime:>9} {rate:>10} "
             "{imbalance:>9}  {idle:>5}").\
             format(explored="Expl",
                    unexplored="Unexpl",
                    objective="Incumbent",
                    bound="Bound",
                    runtime="Time (s)",
                    rgap="Rel. Gap",
                    agap="Abs. Gap",
                    rate="Nodes/Sec",
                    imbalance="Imbalance",
                    idle="Idle")
        self._line_template = \
            ("{tag:>1}{explored:>9d} {unexplored:>9d}  |{objective:>15.7g} "
             "{bound:>15.7g} "+rgap_number_str+"% "+agap_number_str+" |{runtime:>9.1f} {rate:>10.2f} "
             "{imbalance:>8.2f}% {idle:>6d}")
        self._line_template_big_gap = \
            ("{tag:>1}{explored:>9d} {unexplored:>9d}  |{objective:>15.7g} "
             "{bound:>15.7g} "+rgap_label_str+"% "+agap_number_str+" |{runtime:>9.1f} {rate:>10.2f} "
             "{imbalance:>8.2f}% {idle:>6d}")

        self._last_print_time = float('-inf')
        served, explored, unexplored = self._dispatcher._get_node_counts()
        assert served == 0
        assert explored == 0
        assert unexplored == 0
        self._last_explored_nodes_count = 0
        self._last_rgap = None
        self._smoothing = 0.95
        self._avg_time_per_node = None
        self._print_count = 0
        self._new_objective = False
        self._report_new_objective = False

    def log_info(self, msg):
        """Pass a message to ``log.info``"""
        self._log.info(msg)

    def log_warning(self, msg):
        """Pass a message to ``log.warning``"""
        self._log.warning(msg)

    def log_debug(self, msg):
        """Pass a message to ``log.debug``"""
        self._log.debug(msg)

    def log_error(self, msg):
        """Pass a message to ``log.error``"""
        self._log.error(msg)

    def log_critical(self, msg):
        """Pass a message to ``log.critical``"""
        self._log.critical(msg)

    def new_objective(self, report=True):
        """Indicate that a new objective has been found

        Parameters
        ----------
        report : bool, optional
            Indicate whether or not to force the next `tic`
            log output. (default: False)
        """
        self._new_objective = True
        self._report_new_objective = report

    def tic(self, force=False):
        """Provide an opportunity to log output if certain
        criteria are met.

        Parameters
        ----------
        force : bool, optional
            Indicate whether or not to force logging of
            output, even if logging criteria are not
            met. (default: False)
        """
        if not self._log.isEnabledFor(logging.INFO):
            return
        current_time = self._dispatcher.clock()
        new_objective = self._new_objective
        report_new_objective = self._report_new_objective
        delta_t = current_time - self._last_print_time
        self._report_new_objective = False
        if not force:
            if not report_new_objective:
                if delta_t < self._log_interval_seconds:
                    return
        self._new_objective = False
        (served_nodes_count,
         explored_nodes_count,
         unexplored_nodes_count) = self._dispatcher._get_node_counts()
        delta_n = explored_nodes_count - \
                  self._last_explored_nodes_count
        if delta_t and delta_n:
            if self._avg_time_per_node is None:
                self._avg_time_per_node = delta_t / float(delta_n)
            else:
                self._avg_time_per_node = \
                    self._smoothing * delta_t / float(delta_n) + \
                    (1 - self._smoothing) * self._avg_time_per_node
        if self._avg_time_per_node:
            rate = 1.0/self._avg_time_per_node
        else:
            rate = 0.0

        imbalance = self._dispatcher._compute_load_imbalance()
        idle = 0
        if hasattr(self._dispatcher,"needs_work_queue"):
            idle = len(self._dispatcher.needs_work_queue)

        tag = '' if (not new_objective) else '*'
        bound = self._dispatcher._get_current_bound()
        objective = self._dispatcher.best_objective
        agap = self._dispatcher.converger.\
               compute_absolute_gap(bound, objective)
        rgap = self._dispatcher.converger.\
               compute_relative_gap(bound, objective)
        rgap *= 100.0
        if (self._print_count % 5) == 0:
            if self._print_count == 0:
                self._log.info(self._initial_header_line)
            self._log.info(self._header_line)

        if (rgap != inf) and \
           (rgap > 9999.0):
            self._log.info(self._line_template_big_gap.format(
                tag=tag,
                explored=explored_nodes_count,
                unexplored=unexplored_nodes_count,
                objective=objective,
                bound=bound,
                rgap="9999+",
                agap=agap,
                runtime=current_time-self._dispatcher.start_time,
                rate=rate,
                imbalance=imbalance,
                idle=idle))
            self._last_rgap = None
        else:
            self._log.info(self._line_template.format(
                tag=tag,
                explored=explored_nodes_count,
                unexplored=unexplored_nodes_count,
                objective=objective,
                bound=bound,
                rgap=rgap,
                agap=agap,
                runtime=current_time-self._dispatcher.start_time,
                rate=rate,
                imbalance=imbalance,
                idle=idle))
            self._last_rgap = rgap
        self._last_explored_nodes_count = explored_nodes_count
        self._print_count += 1
        self._last_print_time = current_time

class DispatcherBase(object):
    """The base dispatcher implementation with some core
    functionality shared by the distributed and local
    implementations."""

    def __init__(self):
        self.start_time = None
        self.initialized = False
        self.log_new_incumbent = False
        self.termination_condition = None
        self.converger = None
        self.last_global_bound = None
        self.queue = None
        self.best_node = None
        self.best_objective = None
        self.journalist = None
        self.node_limit = None
        self.time_limit = None
        self.served_nodes_count = None
        self.worst_terminal_bound = None
        self.clock = None

    def _add_work_to_queue(self, node):
        bound = node.bound
        if self.converger.eligible_for_queue(
                bound,
                self.best_objective):
            self.queue.put(node)
            return True
        else:
            self._check_update_worst_terminal_bound(bound)
            return False

    def _check_update_worst_terminal_bound(self, bound):
        if (self.worst_terminal_bound is None) or \
           self.converger.bound_worsened(bound,
                                         self.worst_terminal_bound):
            self.worst_terminal_bound = bound

    def _check_update_best_objective(self, objective):
        assert objective is not None
        assert not math.isnan(objective)
        updated = False
        if self.converger.objective_improved(
                objective,
                self.best_objective):
            self.best_objective = objective
            updated = True
            if self.journalist is not None:
                self.journalist.new_objective(
                    report=self.log_new_incumbent)
            eligible_for_queue_ = self.converger.eligible_for_queue
            removed = self.queue.filter(
                lambda node_: eligible_for_queue_(
                    node_.bound,
                    self.best_objective))
            for node_ in removed:
                self._check_update_worst_terminal_bound(node_.bound)
        return updated

    def _check_update_best_node(self, node):
        best_objective_updated = False
        objective = node.objective
        assert objective is not None
        assert not math.isnan(objective)
        if (objective != self.converger.infeasible_objective) and \
           ((self.best_node is None) or \
            self.converger.objective_improved(
                objective,
                self.best_node.objective)):
            assert node._uuid is not None
            self.best_node = node
            best_objective_updated = \
                self._check_update_best_objective(objective)
        return best_objective_updated

    def _check_convergence(self):
        # check if we are done
        if self.termination_condition is None:
            global_bound = self._get_current_bound()
            self.last_global_bound = global_bound
            self.termination_condition = self.converger.\
                check_termination_criteria(global_bound,
                                           self.best_objective)
            if self.termination_condition is None:
                if self.node_limit is not None:
                    (served_nodes_count, explored_nodes_count, _) = \
                        self._get_node_counts()
                    # we need to check the explored count as
                    # it may be greater than the served
                    # count due to the use of a nested solver
                    if (served_nodes_count >= self.node_limit) or \
                       (explored_nodes_count >= self.node_limit):
                        self.termination_condition = \
                            TerminationCondition.node_limit
                if self.termination_condition is None:
                    if (self.time_limit is not None) and \
                       ((self.clock() - self.start_time) >= self.time_limit):
                        self.termination_condition = \
                            TerminationCondition.time_limit

    def _get_work_item(self):
        node = self.queue.get()
        assert node is not None
        self.served_nodes_count += 1
        return node

    #
    # Interface
    #

    def initialize(self,
                   best_objective,
                   best_node,
                   initialize_queue,
                   queue_strategy,
                   converger,
                   node_limit,
                   time_limit,
                   log,
                   log_interval_seconds,
                   log_new_incumbent):
        """Initialize the dispatcher for a new solve.

        Parameters
        ----------
        best_objective : float
            The assumed best objective to start with.
        best_node : :class:`Node <pybnb.node.Node>`
            A node storing the assumed best objective.
        initialize_queue : :class:`pybnb.dispatcher.DispatcherQueueData`
            The initial queue.
        queue_strategy : :class:`QueueStrategy <pybnb.common.QueueStrategy>`
            Sets the strategy for prioritizing nodes in the
            central dispatcher queue. See the
            :class:`QueueStrategy <pybnb.common.QueueStrategy>`
            enum for the list of acceptable values.
        converger : :class:`pybnb.convergence_checker.ConvergenceChecker`
            The branch-and-bound convergence checker object.
        node_limit : int or None
            An integer representing the maximum number of
            nodes to processes before beginning to terminate
            the solve. If None, no node limit will be
            enforced.
        time_limit : float or None
            The maximum amount of time to spend processing
            nodes before beginning to terminate the
            solve. If None, no time limit will be enforced.
        log : ``logging.Logger``
            A log object where solver output should be sent.
        log_interval_seconds : float
            The approximate maximum time (in seconds)
            between solver log updates. More time may pass
            between log updates if no updates have been
            received from any workers, and less time may
            pass if a new incumbent is found.
        log_new_incumbent : bool
            Controls whether updates to the best objective
            are logged immediately (overriding the log
            interval). Setting this to false can be useful
            when frequent updates to the incumbent are
            expected and the additional logging slows down
            the dispatcher.
        """
        assert (node_limit is None) or \
            ((node_limit > 0) and \
             (node_limit == int(node_limit)))
        assert (time_limit is None) or \
            (time_limit >= 0)
        self.start_time = self.clock()
        self.initialized = True
        self.log_new_incumbent = log_new_incumbent
        self.termination_condition = None
        self.converger = converger
        if initialize_queue.sense != self.converger.sense:
            raise ValueError("The objective sense does not match "
                             "that of the initial queue.")
        self.last_global_bound = self.converger.unbounded_objective
        self.queue = PriorityQueueFactory(queue_strategy,
                                          self.converger.sense)
        self.best_objective = best_objective
        self.best_node = None
        if best_node is not None:
            assert best_node._uuid is not None
            self._check_update_best_node(best_node)
        self.node_limit = None
        if node_limit is not None:
            self.node_limit = int(node_limit)
        self.time_limit = None
        if time_limit is not None:
            self.time_limit = float(time_limit)
        self.served_nodes_count = 0
        self.worst_terminal_bound = \
            initialize_queue.worst_terminal_bound
        self.journalist = None
        if (log is not None) and (not log.disabled):
            self.journalist = StatusPrinter(
                self,
                log,
                log_interval_seconds=log_interval_seconds)
        for node in initialize_queue.nodes:
            self._add_work_to_queue(node)

    def log_info(self, msg):
        """Pass a message to ``log.info``"""
        if self.journalist is not None:
            self.journalist.log_info(msg)

    def log_warning(self, msg):
        """Pass a message to ``log.warning``"""
        if self.journalist is not None:
            self.journalist.log_warning(msg)

    def log_debug(self, msg):
        """Pass a message to ``log.debug``"""
        if self.journalist is not None:
            self.journalist.log_debug(msg)

    def log_error(self, msg):
        """Pass a message to ``log.error``"""
        if self.journalist is not None:
            self.journalist.log_error(msg)

    def log_critical(self, msg):
        """Pass a message to ``log.critical``"""
        if self.journalist is not None:
            self.journalist.log_critical(msg)

    def save_dispatcher_queue(self):
        """Saves the current dispatcher queue. The result can
        be used to re-initialize a solve.

        Returns
        -------
        queue_data : :class:`pybnb.dispatcher.DispatcherQueueData`
            An object storing information that can be used
            to re-initialize the dispatcher queue to its
            current state.
        """
        return DispatcherQueueData(
            nodes=list(self.queue.items()),
            worst_terminal_bound=self.worst_terminal_bound,
            sense=self.converger.sense)

    #
    # Abstract Methods
    #

    def update(self, *args, **kwds):              #pragma:nocover
        raise NotImplementedError

    def _compute_load_imbalance(self):            #pragma:nocover
        """Get the worker load imbalance."""
        raise NotImplementedError()

    def _get_current_bound(self):                 #pragma:nocover
        """Get the current global bound"""
        raise NotImplementedError()

    def _get_final_solve_info(self):              #pragma:nocover
        """Get the final solve information"""
        raise NotImplementedError()

    def _get_node_counts(self):                   #pragma:nocover
        """Get the served and explored node counts"""
        raise NotImplementedError()

class DispatcherLocal(DispatcherBase):
    """The central dispatcher for a serial branch-and-bound
    algorithm."""

    def __init__(self):
        super(DispatcherLocal, self).__init__()
        self.external_bound = None
        self.solve_info = _SolveInfo()
        self.active_nodes = 0
        self.clock = time.time

    def _compute_load_imbalance(self):
        """Get the worker load imbalance."""
        return 0.0

    def _get_current_bound(self):
        """Get the current global bound"""
        bound = self.queue.bound()
        if self.converger.sense == maximize:
            if (self.external_bound is not None) and \
               ((bound is None) or \
                (self.external_bound > bound)):
                bound = self.external_bound
            if (self.worst_terminal_bound is not None) and \
               ((bound is None) or \
                (self.worst_terminal_bound > bound)):
                bound = self.worst_terminal_bound
        else:
            if (self.external_bound is not None) and \
               ((bound is None) or \
                (self.external_bound < bound)):
                bound = self.external_bound
            if (self.worst_terminal_bound is not None) and \
               ((bound is None) or \
                (self.worst_terminal_bound < bound)):
                bound = self.worst_terminal_bound
        if bound is not None:
            return bound
        else:
            # likely means the dispatcher was initialized
            # with an empty queue
            return self.converger.unbounded_objective

    def _get_final_solve_info(self):
        """Get the final solve information"""
        solve_info = _SolveInfo()
        solve_info.data[:] = self.solve_info.data
        return solve_info

    def _get_node_counts(self):
        return (self.served_nodes_count,
                self.solve_info.explored_nodes_count,
                self.queue.size() + self.active_nodes)

    #
    # Overloaded base class methods
    #

    def _check_update_best_objective(self, objective):
        updated = super(DispatcherLocal, self).\
            _check_update_best_objective(objective)
        if updated:
            assert self.best_objective == objective
            if self.external_bound is not None:
                if not self.converger.eligible_for_queue(
                        self.external_bound,
                        objective):
                    self.external_bound = None

    #
    # Interface
    #

    def initialize(self,
                   best_objective,
                   best_node,
                   initialize_queue,
                   queue_strategy,
                   converger,
                   node_limit,
                   time_limit,
                   log,
                   log_interval_seconds,
                   log_new_incumbent):
        """Initialize the dispatcher. See the
        :func:`pybnb.dispatcher.DispatcherBase.initialize`
        method for argument descriptions."""
        self.solve_info.reset()
        self.active_nodes = 0
        super(DispatcherLocal, self).initialize(
            best_objective,
            best_node,
            initialize_queue,
            queue_strategy,
            converger,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        if self.journalist is not None:
            self.log_info("Starting branch & bound solve:\n"
                          " - dispatcher pid: %s (%s)\n"
                          " - worker processes: 1"
                          % (os.getpid(),
                             socket.gethostname()))
            self.journalist.tic()

    def update(self,
               best_objective,
               best_node,
               terminal_bound,
               solve_info,
               node_list):
        """Update local worker information.

        Parameters
        ----------
        best_objective : float or None
            A new potential best objective found by the
            worker.
        best_node : :class:`Node <pybnb.node.Node>` or None
            A new potential best node found by the worker.
        terminal_bound : float or None
            The worst bound of any terminal nodes that were
            processed by the worker since the last update.
        solve_info : :class:`_SolveInfo`
            The most up-to-date worker solve information.
        node_list : list
            A list of nodes to add to the queue.

        Returns
        -------
        solve_finished : bool
            Indicates if the dispatcher has terminated the solve.
        new_objective : float
            The best objective known to the dispatcher.
        best_node : :class:`Node <pybnb.node.Node>` or None
            The best node known to the dispatcher.
        data : :class:`Node <pybnb.node.Node>` or None
            If solve_finished is false, a new node for the
            worker to process. Otherwise, a tuple containing
            the global bound, the termination condition
            string, and the number of explored nodes.
        """
        assert self.initialized
        if best_objective is not None:
            self._check_update_best_objective(
                best_objective)
        if best_node is not None:
            assert best_node._uuid is not None
            self._check_update_best_node(best_node)
        self.solve_info.data[:] = solve_info.data
        self.external_bound = None
        self.active_nodes = 0
        if len(node_list) > 0:
            for node in node_list:
                self._add_work_to_queue(node)
        if terminal_bound is not None:
            self._check_update_worst_terminal_bound(
                    terminal_bound)
        last_global_bound = self.last_global_bound
        self._check_convergence()
        if (self.queue.size() == 0) and \
           (self.termination_condition is None):
            self.termination_condition = \
                TerminationCondition.no_nodes
        if self.termination_condition is None:
            node = self._get_work_item()
            self.active_nodes = 1
            self.external_bound = node.bound
            if self.journalist is not None:
                force = (last_global_bound == \
                         self.converger.unbounded_objective) and \
                         (last_global_bound != \
                          self.last_global_bound)
                self.journalist.tic(force=force)
            return (False,
                    self.best_objective,
                    self.best_node,
                    node)
        else:
            if self.journalist is not None:
                self.journalist.tic(force=True)
                self.journalist.log_info(self.journalist._lines)
            self.initialized = False
            return (True,
                    self.best_objective,
                    self.best_node,
                    (self._get_current_bound(),
                     self.termination_condition,
                     self._get_final_solve_info()))

class DispatcherDistributed(DispatcherBase):
    """The central dispatcher for a distributed
    branch-and-bound algorithm.

    Parameters
    ----------
    comm : ``mpi4py.MPI.Comm``, optional
        The MPI communicator to use. If set to None, this
        will disable the use of MPI and avoid an attempted
        import of `mpi4py.MPI` (which avoids triggering a
        call to `MPI_Init()`).
    """

    def __init__(self, comm):
        assert comm.size > 1
        import mpi4py.MPI
        assert mpi4py.MPI.Is_initialized()
        super(DispatcherDistributed, self).__init__()
        self.clock = mpi4py.MPI.Wtime
        self.comm = comm
        # send rank of dispatcher to all workers
        self.dispatcher_rank = DispatcherProxy._init(
            self.comm,
            ProcessType.dispatcher)
        assert self.dispatcher_rank == self.comm.rank
        self.worker_ranks = [i for i in range(self.comm.size)
                             if i != self.comm.rank]
        self.needs_work_queue = \
            collections.deque([],
                              len(self.worker_ranks))
        self._solve_info_by_source = \
            {i: _SolveInfo() for i in self.worker_ranks}
        self.last_known_bound = dict()
        self.external_bounds = SortedList()
        self.has_work = set()
        self._send_requests = None
        self.explored_nodes_count = 0

    def _compute_load_imbalance(self):
        pmin = inf
        pmax = -inf
        psum = 0
        for solve_info in self._solve_info_by_source.values():
            count = solve_info.explored_nodes_count
            if count < pmin:
                pmin = count
            if count > pmax:
                pmax = count
            psum += count
        if psum > 0:
            pavg = psum/float(len(self._solve_info_by_source))
            return (pmax-pmin)/pavg*100.0
        else:
            return 0.0

    def _get_current_bound(self):
        """Get the current global bound"""
        bound = self.queue.bound()
        if self.converger.sense == maximize:
            if len(self.external_bounds) and \
               ((bound is None) or \
                (self.external_bounds[-1] > bound)):
                bound = self.external_bounds[-1]
            if (self.worst_terminal_bound is not None) and \
               ((bound is None) or \
                (self.worst_terminal_bound > bound)):
                bound = self.worst_terminal_bound
        else:
            if len(self.external_bounds) and \
               ((bound is None) or \
                (self.external_bounds[0] < bound)):
                bound = self.external_bounds[0]
            if (self.worst_terminal_bound is not None) and \
               ((bound is None) or \
                (self.worst_terminal_bound < bound)):
                bound = self.worst_terminal_bound
        if bound is not None:
            return bound
        else:
            # likely means the dispatcher was initialized
            # with an empty queue
            return self.converger.unbounded_objective

    def _get_final_solve_info(self):
        solve_info = _SolveInfo()
        for worker_solve_info in self._solve_info_by_source.values():
            solve_info.add_from(worker_solve_info)
        return solve_info

    def _get_node_counts(self):
        return (self.served_nodes_count,
                self.explored_nodes_count,
                self.queue.size() + len(self.has_work))

    #
    # Overloaded base class methods
    #

    def _check_update_best_objective(self, objective):
        updated = super(DispatcherDistributed, self).\
            _check_update_best_objective(objective)
        if updated:
            assert self.best_objective == objective
            self_external_bounds = self.external_bounds
            eligible_for_queue = self.converger.eligible_for_queue
            # trim the sorted external_bounds list
            N = len(self_external_bounds)
            if self.converger.sense == maximize:
                i = 0
                for i in range(N):
                    if eligible_for_queue(
                            self_external_bounds[i],
                            objective):
                        break
                if i != 0:
                    self.external_bounds = SortedList(
                        self_external_bounds.islice(i, N))
            else:
                i = N-1
                for i in range(N-1,-1,-1):
                    if eligible_for_queue(
                            self_external_bounds[i],
                            objective):
                        break
                if i != N-1:
                    self.external_bounds = SortedList(
                        self_external_bounds.islice(0, i+1))

    def _get_work_to_send(self, dest):
        node = self._get_work_item()
        bound = node.bound
        self.last_known_bound[dest] = bound
        self.external_bounds.add(bound)
        self.has_work.add(dest)
        return node

    def _send_work(self):
        stop = False
        data = None
        if len(self.needs_work_queue) > 0:
            if self._send_requests is None:
                self._send_requests = \
                    {i: None for i in self.worker_ranks}
            if self.termination_condition is None:
                while (self.queue.size() > 0) and \
                      (len(self.needs_work_queue) > 0):
                    stop = False
                    dest = self.needs_work_queue.popleft()
                    node = self._get_work_to_send(dest)
                    best_node_slots = None
                    if self.best_node is not None:
                        best_node_slots = self.best_node.slots
                    send_ = marshal.dumps(
                        (self.best_objective,
                         best_node_slots,
                         node.slots),
                        config.MARSHAL_PROTOCOL_VERSION)
                    if self._send_requests[dest] is not None:
                        self._send_requests[dest].Wait()
                    self._send_requests[dest] = \
                        self.comm.Isend([send_,mpi4py.MPI.BYTE],
                                        dest,
                                        tag=DispatcherResponse.work)
                    # a shortcut to check if we should keep sending nodes
                    if (self.node_limit is not None) and \
                       (self.served_nodes_count >= self.node_limit):
                        break
            if len(self.needs_work_queue) == (self.comm.size-1):
                if self.termination_condition is None:
                    self.termination_condition = \
                        TerminationCondition.no_nodes
                requests = []
                for r_ in self._send_requests.values():
                    if r_ is not None:
                        requests.append(r_)
                mpi4py.MPI.Request.Waitall(requests)
                self._send_requests = None
                stop = True
                data = (self._get_current_bound(),
                        self.termination_condition,
                        self._get_final_solve_info())
                best_node_slots = None
                if self.best_node is not None:
                    best_node_slots = self.best_node.slots
                send_ = marshal.dumps(
                    (self.best_objective,
                     best_node_slots,
                     data[0],
                     _termination_condition_to_int[data[1]],
                     data[2].data),
                    config.MARSHAL_PROTOCOL_VERSION)
                # everyone needs work, so we must be done
                requests = []
                while len(self.needs_work_queue) > 0:
                    dest = self.needs_work_queue.popleft()
                    requests.append(
                        self.comm.Isend([send_,mpi4py.MPI.BYTE],
                                        dest,
                                        DispatcherResponse.nowork))
                mpi4py.MPI.Request.Waitall(requests)

        return (stop, self.best_objective, self.best_node, data)

    def _update_solve_info(self, solve_info_data, source):
        self.explored_nodes_count -= \
            self._solve_info_by_source[source].explored_nodes_count
        self._solve_info_by_source[source].data[:] = solve_info_data
        self.explored_nodes_count += \
            self._solve_info_by_source[source].explored_nodes_count

    #
    # Interface
    #

    def initialize(self,
                   best_objective,
                   best_node,
                   initialize_queue,
                   queue_strategy,
                   converger,
                   node_limit,
                   time_limit,
                   log,
                   log_interval_seconds,
                   log_new_incumbent):
        """Initialize the dispatcher. See the
        :func:`pybnb.dispatcher.DispatcherBase.initialize`
        method for argument descriptions."""
        self.needs_work_queue.clear()
        for solve_info in self._solve_info_by_source.values():
            solve_info.reset()
        self.last_known_bound.clear()
        self.external_bounds.clear()
        self.has_work.clear()
        self._send_requests = None
        self.explored_nodes_count = 0
        if best_node is not None:
            best_node = _SerializedNode.from_node(best_node)
        initialize_queue_ = DispatcherQueueData(
            nodes=[_SerializedNode.from_node(node) \
                   if (type(node) is not _SerializedNode) else node
                   for node in initialize_queue.nodes],
            worst_terminal_bound=initialize_queue.worst_terminal_bound,
            sense=initialize_queue.sense)
        super(DispatcherDistributed, self).initialize(
            best_objective,
            best_node,
            initialize_queue_,
            queue_strategy,
            converger,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        if self.journalist is not None:
            self.log_info("Starting branch & bound solve:\n"
                          " - dispatcher pid: %s (%s)\n"
                          " - worker processes: %d"
                          % (os.getpid(),
                             socket.gethostname(),
                             len(self.worker_ranks)))
            self.journalist.tic()

    def update(self,
               best_objective,
               best_node,
               terminal_bound,
               solve_info,
               node_list,
               source):
        """Update local worker information.

        Parameters
        ----------
        best_objective : float or None
            A new potential best objective found by the
            worker.
        best_node : :class:`Node <pybnb.node.Node>` or None
            A new potential best node found by the worker.
        terminal_bound : float or None
            The worst bound of any terminal nodes that were
            processed by the worker since the last update.
        solve_info : :class:`_SolveInfo`
            The most up-to-date worker solve information.
        node_list : list
            A list of nodes to add to the queue.
        source : int
            The worker process rank that the update came from.

        Returns
        -------
        solve_finished : bool
            Indicates if the dispatcher has terminated the solve.
        new_objective : float
            The best objective value known to the dispatcher.
        best_node : :class:`Node <pybnb.node.Node>` or None
            The best node known to the dispatcher.
        data : ``array.array`` or None
            If solve_finished is false, a data array
            representing a new node for the worker to
            process. Otherwise, a tuple containing the
            global bound, the termination condition string,
            and the number of explored nodes.
        """
        assert self.initialized
        self._update_solve_info(
            solve_info.data,
            source)
        self.needs_work_queue.append(source)
        self.has_work.discard(source)
        if source in self.last_known_bound:
            val_ = self.last_known_bound[source]
            try:
                self.external_bounds.remove(val_)
            except ValueError:
                # rare, but can happen when
                # _check_update_best_node modifies
                # the external_bounds list
                pass
        if best_objective is not None:
            self._check_update_best_objective(
                best_objective)
        if best_node is not None:
            assert best_node._uuid is not None
            self._check_update_best_node(best_node)
        if len(node_list) > 0:
            for node in node_list:
                self._add_work_to_queue(node)
        if terminal_bound is not None:
            self._check_update_worst_terminal_bound(
                terminal_bound)
        last_global_bound = self.last_global_bound
        self._check_convergence()
        ret = self._send_work()
        stop = ret[0]
        if not stop:
            if self.journalist is not None:
                force = (last_global_bound == \
                         self.converger.unbounded_objective) and \
                         (last_global_bound != \
                          self.last_global_bound)
                self.journalist.tic(force=force)
        else:
            if self.journalist is not None:
                self.journalist.tic(force=True)
                self.journalist.log_info(self.journalist._lines)
            assert self.initialized
            self.initialized = False
        return ret

    #
    # Distributed Interface
    #

    def serve(self):
        """Start listening for distributed branch-and-bound
        commands and map them to commands in the local
        dispatcher interface."""

        def rebuild_update_requests(size):
            update_requests = {}
            update_data = bytearray(size)
            for i in self.worker_ranks:
                update_requests[i] = self.comm.Recv_init(
                    update_data,
                    source=i,
                    tag=DispatcherAction.update)
            return update_requests, update_data

        update_requests = None
        solve_info_ = _SolveInfo()
        data = None
        msg = Message(self.comm)
        while (1):
            msg.probe()
            tag = msg.tag
            source = msg.source
            if tag == DispatcherAction.update:
                size = msg.status.Get_count(datatype=mpi4py.MPI.BYTE)
                if (data is None) or \
                   (len(data) < size):
                    update_requests, data = \
                        rebuild_update_requests(size)
                req = update_requests[msg.status.Get_source()]
                req.Start()
                req.Wait()
                if six.PY2:
                    data_ = str(data)
                else:
                    data_ = data
                (best_objective,
                 best_node,
                 terminal_bound,
                 solve_info_data,
                 node_list) = marshal.loads(data_)
                solve_info_.data = array.array('d',solve_info_data)
                if best_node is not None:
                    best_node = _SerializedNode(best_node)
                node_list = [_SerializedNode(state)
                             for state in node_list]
                ret = self.update(best_objective,
                                  best_node,
                                  terminal_bound,
                                  solve_info_,
                                  node_list,
                                  source)
                stop = ret[0]
                if stop:
                    best_node = ret[2]
                    if best_node is not None:
                        best_node = _SerializedNode.restore_node(
                            best_node.slots)
                    return (ret[1],     # best_objective
                            best_node,
                            ret[3][0],  # global_bound
                            ret[3][1],  # termination_condition
                            ret[3][2])  # global_solve_info
            elif tag == DispatcherAction.log_info:
                msg.recv(mpi4py.MPI.CHAR)
                self.log_info(msg.data)
            elif tag == DispatcherAction.log_warning:
                msg.recv(mpi4py.MPI.CHAR)
                self.log_warning(msg.data)
            elif tag == DispatcherAction.log_debug:
                msg.recv(mpi4py.MPI.CHAR)
                self.log_debug(msg.data)
            elif tag == DispatcherAction.log_error:
                msg.recv(mpi4py.MPI.CHAR)
                self.log_error(msg.data)
            elif tag == DispatcherAction.log_critical:
                msg.recv(mpi4py.MPI.CHAR)
                self.log_critical(msg.data)
            elif tag == DispatcherAction.stop_listen:
                msg.recv()
                assert msg.data is None
                return (None, None, None, None, None)
            else:                                 #pragma:nocover
                raise RuntimeError("Dispatcher received invalid "
                                   "message tag '%s' from rank '%s'"
                                   % (tag, source))

    def save_dispatcher_queue(self):
        """Saves the current dispatcher queue. The result can
        be used to re-initialize a solve.

        Returns
        -------
        queue_data : :class:`pybnb.dispatcher.DispatcherQueueData`
            An object storing information that can be used
            to re-initialize the dispatcher queue to its
            current state.
        """
        nodes = []
        for node in self.queue.items():
            node_ = node.restore_node(node.slots)
            nodes.append(node_)
        return DispatcherQueueData(
            nodes=nodes,
            worst_terminal_bound=self.worst_terminal_bound,
            sense=self.converger.sense)



import os
import tempfile
import time

import pytest

from pybnb.common import (minimize,
                          inf,
                          nan)
from pybnb.node import Node
from pybnb.problem import Problem
from pybnb.solver import (Solver,
                          summarize_worker_statistics,
                          solve)

from six import StringIO

class BadBranchSignatureProblem(Problem):
    def sense(self): return minimize
    def objective(self): return 0
    def bound(self): return 0
    def save_state(self, node): pass
    def load_state(self, node): pass
    def branch(self, node): raise NotImplementedError()

class DummyProblem(Problem):
    def sense(self): return minimize
    def objective(self): return 0
    def bound(self): return 0
    def save_state(self, node): pass
    def load_state(self, node): pass
    def branch(self): raise NotImplementedError()

class _DummyComm_Size1(object):
    size = 1

class TestSolverSimple(object):

    def test_bad_dispatcher_rank(self):
        with pytest.raises(ValueError):
            Solver(comm=None, dispatcher_rank=-1)
        with pytest.raises(ValueError):
            Solver(comm=None, dispatcher_rank=1)
        with pytest.raises(ValueError):
            Solver(comm=None, dispatcher_rank=1.1)
        Solver(comm=None, dispatcher_rank=0)
        Solver(comm=None)

    def test_no_mpi(self):
        b = Solver(comm=None)
        assert b.comm is None
        assert b.worker_comm is None
        assert b.is_worker == True
        assert b.is_dispatcher == True
        assert b.worker_count == 1
        b._reset_local_solve_stats()
        stats = b.collect_worker_statistics()
        assert len(stats) == 12
        assert stats['wall_time'] == [0]
        assert stats['queue_time'] == [0]
        assert stats['queue_call_count'] == [0]
        assert stats['objective_time'] == [0]
        assert stats['objective_call_count'] == [0]
        assert stats['bound_time'] == [0]
        assert stats['bound_call_count'] == [0]
        assert stats['branch_time'] == [0]
        assert stats['branch_call_count'] == [0]
        assert stats['load_state_time'] == [0]
        assert stats['load_state_call_count'] == [0]
        assert stats['explored_nodes_count'] == [0]
        out = \
"""Number of Workers:        1
Load Imbalance:       0.00%
Average Worker Timing:
 - queue:       0.00% [avg time:   0.0 s , count: 0]
 - load_state:  0.00% [avg time:   0.0 s , count: 0]
 - bound:       0.00% [avg time:   0.0 s , count: 0]
 - objective:   0.00% [avg time:   0.0 s , count: 0]
 - branch:      0.00% [avg time:   0.0 s , count: 0]
 - other:       0.00% [avg time:   0.0 s , count: 0]
"""
        tmp = StringIO()
        summarize_worker_statistics(stats, stream=tmp)
        assert tmp.getvalue() == out

    def test_solve_function(self):
        fid, fname = tempfile.mkstemp()
        os.close(fid)
        try:
            solve(DummyProblem(),
                  comm=None,
                  log_filename=fname)
            assert os.path.exists(fname)
        finally:
            time.sleep(0.1)
            try:
                os.remove(fname)
            except:                               #pragma:nocover
                pass

    def test_bad_queue_strategy(self):
        with pytest.raises(ValueError):
            solve(DummyProblem(),
                  comm=None,
                  queue_strategy='_not_a_valid_strategy_')

    def test_bad_best_options(self):
        node = Node()
        node.objective = None
        with pytest.raises(ValueError):
            solve(DummyProblem(),
                  comm=None,
                  best_node=node)
        node.objective = nan
        with pytest.raises(ValueError):
            solve(DummyProblem(),
                  comm=None,
                  best_node=node)
        node.objective = 0
        solve(DummyProblem(),
              comm=None,
              best_node=node)

    def test_bad_branch_signature(self):
        problem = BadBranchSignatureProblem()
        with pytest.raises(TypeError):
            solve(problem,
                  comm=None)



"""
Miscellaneous utilities used for development.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import logging
import signal

class MPI_InterruptHandler(object):
    """A context manager for temporarily assigning a handler
    to SIGINT and SIGUSR1, depending on the availability of
    these signals in the current OS."""
    _sigs = [signal.SIGINT]
    if hasattr(signal, 'SIGUSR1'):
        # not available on windows
        _sigs.append(signal.SIGUSR1)
    __slots__ = ("_released",
                 "_original_handlers",
                 "_handler",
                 "_disable")
    def __init__(self, handler, disable=False):
        self._released = True
        self._original_handlers = None
        self._handler = handler
        self._disable = disable

    def __enter__(self):
        if self._disable:
            return self
        self._released = False
        self._original_handlers = \
            [(signum, signal.getsignal(signum))
             for signum in self._sigs]
        def handler(signum, frame):
            self._handler(signum, frame)
            self.release()
        for signum in self._sigs:
            signal.signal(signum, handler)
        return self

    def __exit__(self, type, value, tb):
        self.release()

    def release(self):
        if not self._released:
            for signum, handler in self._original_handlers:
                signal.signal(signum, handler)
            self._released = True

def metric_format(num, unit="s", digits=1, align_unit=False):
    """Format and scale output with metric prefixes.

    Example
    -------

    >>> metric_format(0)
    '0.0 s'
    >>> metric_format(0, align_unit=True)
    '0.0 s '
    >>> metric_format(0.002, unit='B')
    '2.0 mB'
    >>> metric_format(2001, unit='B')
    '2.0 KB'
    >>> metric_format(2001, unit='B', digits=3)
    '2.001 KB'

    """
    if num is None:
        return "<unknown>"
    prefix = ""
    if (num >= 1.0) or (num == 0.0):
        if num >= 1000.0:
            num /= 1000.0
            for p in ['K','M','G','T','P','E','Z','Yi']:
                prefix = p
                if abs(num) < 1000.0:
                    break
                num /= 1000.0
    else:
        num *= 1000.0
        for p in ['m','u','n','p','f']:
            prefix = p
            if abs(num) > 1:
                break
            num *= 1000.0
    if (prefix == "") and align_unit:
        return ("%."+str(digits)+"f %s ") % (num, unit)
    else:
        return ("%."+str(digits)+"f %s%s") % (num, prefix, unit)

def time_format(num, digits=1, align_unit=False):
    """Format and scale output according to standard time
    units.

    Example
    -------

    >>> time_format(0)
    '0.0 s'
    >>> time_format(0, align_unit=True)
    '0.0 s '
    >>> time_format(0.002)
    '2.0 ms'
    >>> time_format(2001)
    '33.4 m'
    >>> time_format(2001, digits=3)
    '33.350 m'

    """
    if num is None:
        return "<unknown>"
    unit = "s"
    if (num >= 1.0) or (num == 0.0):
        if num >= 60.0:
            num /= 60.0
            unit = "m"
            if num >= 60.0:
                num /= 60.0
                unit = "h"
                if num >= 24.0:
                    num /= 24.0
                    unit = "d"
    else:
        num *= 1000.0
        for p in ['ms','us','ns','ps','fs']:
            unit = p
            if abs(num) > 1:
                break
            num *= 1000.0
    if (len(unit) == 1) and align_unit:
        return ("%."+str(digits)+"f %s ") % (num, unit)
    else:
        return ("%."+str(digits)+"f %s") % (num, unit)

def get_gap_labels(gap,
                   key="gap",
                   format="f"):
    """Get format strings with enough size and precision to print
    a given gap tolerance."""
    gap_length = 10
    gap_digits = 0
    while gap < (10**(-gap_digits+1)):
        gap_digits += 1
        if gap_length - gap_digits < 5:
            gap_length += 1
    gap_label_str = "{"+key+":>"+str(gap_length)+"}"
    gap_number_str = "{"+key+":>"+str(gap_length)+"." + \
                     str(gap_digits)+format+"}"
    return gap_length, gap_label_str, gap_number_str

class _NullCM(object):
    """A context manager that does nothing"""
    def __init__(self, obj):
        self.obj = obj
    def __enter__(self):
        return self.obj
    def __exit__(self, *args):
        pass

def as_stream(stream, **kwds):
    """A utility for handling function arguments that can be
    a filename or a file object. This function is meant to be
    used in the context of a with statement.

    Parameters
    ----------
    stream : file-like object or string
        An existing file-like object or the name of a file
        to open.
    **kwds
        Additional keywords passed to the built-in function
        ``open`` when the `stream` keyword is a filename.

    Returns
    -------
    file-like object
        A file-like object that can be written to. If the
        input argument was originally an open file, a dummy
        context will wrap the file object so that it will
        not be closed upon exit of the with block.

    Example
    -------

    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile() as f:
    ...     # pass a file
    ...     with as_stream(f) as g:
    ...         assert g is f
    ...     assert not f.closed
    ...     f.close()
    ...     # pass a filename
    ...     with as_stream(f.name) as g:
    ...         assert not g.closed
    ...     assert g.closed

    """
    import six
    if isinstance(stream, six.string_types):
        return open(stream,"w")
    else:
        return _NullCM(stream)

def get_default_args(func):
    """Get the default arguments for a function as a
    dictionary mapping argument name to default value.

    Example
    -------

    >>> def f(a, b=None):
    ...     pass
    >>> get_default_args(f)
    {'b': None}

    """
    import inspect
    import six
    if six.PY3:
        signature = inspect.signature(func)
        return {k: v.default
                for k, v in signature.parameters.items()
                if v.default is not inspect.Parameter.empty}
    else:
        a = inspect.getargspec(func)
        if a.defaults is None:
            return {}
        return dict(zip(a.args[-len(a.defaults):],a.defaults))

def get_keyword_docs(doc):
    """Parses a numpy-style docstring to summarize
    information in the 'Parameters' section into a dictionary."""
    import re
    import yaml
    lines = doc.splitlines()
    for i_start, line in enumerate(lines):
        if (line.strip() == "Parameters") and \
           (lines[i_start+1].strip() == "----------"):
            i_start = i_start + 2
            break
    else:                                              #pragma:nocover
        assert False
    for i_stop, line in enumerate(lines[i_start:],i_start):
        if i_stop <= i_start+1:
            continue
        if line.strip() == "":
            break
    else:                                              #pragma:nocover
        assert False

    args = {}
    choices = {}
    i = i_start
    while i != i_stop:
        if i == i_start:
            assert re.match(r".+ : .+(, optional)?", lines[i])
        if re.match(r".+ : .+(, optional)?", lines[i]):
            if i != i_start:
                args[last] = args[last].strip()
            last = lines[i].split(' : ')[0].strip()
            if re.match(r".+ : \{.+\}(, optional)?", lines[i]):
                assert lines[i].count("{") == 1
                assert lines[i].count("}") == 1
                opts = lines[i].split("{")[1].split("}")[0]
                opts = [eval(c.strip()) for c in opts.split(",")]
                choices[last] = opts
            args[last] = ""
        else:
            args[last] += (lines[i].strip() + " ")
        i += 1
    args[last] = args[last].strip()
    data = {}
    for key, val in args.items():
        data[key] = {"doc": val}
        default_ = re.search(r"\(default: .*\)",val)
        if default_ is not None:
            default_ = default_.group(0)[1:-1].split(": ")
            assert len(default_) == 2
            default_ = eval(default_[1])
            data[key]["default"] = default_
            doc = re.split(r"\(default: .*\)",val)
            assert len(doc) == 2
            assert doc[1].strip() == ""
            data[key]["doc"] = doc[0].strip()
        if key in choices:
            data[key]["choices"] = choices[key]

    return data

class _simple_stdout_filter(object):
    def filter(self, record):
        # only show WARNING or below
        return record.levelno <= logging.WARNING

class _simple_stderr_filter(object):
    def filter(self, record):
        # only show ERROR or above
        return record.levelno >= logging.ERROR

def get_simple_logger(filename=None,
                      stream=None,
                      console=True,
                      level=logging.INFO,
                      formatter=None):
    """Creates a logging object configured to write to any
    combination of a file, a stream, and the console, or
    hide all output.

    Parameters
    ----------
    filename : string, optional
        The name of a file to write to. (default: None)
    stream : file-like object, optional
        A file-like object to write to. (default: None)
    console : bool, optional
        If True, the logger will be configured to print
        output to the console through stdout and
        stderr. (default: True)
    level : int, optional
        The logging level to use. (default: ``logging.INFO``)
    formatter: ``logging.Formatter``, optional
        The logging formatter to use. (default: None)

    Returns
    -------
    ``logging.Logger``
        A logging object
    """
    log = logging.Logger(None, level=level)
    if filename is not None:
        # create file handler which logs even debug messages
        fh = logging.FileHandler(filename)
        fh.setLevel(level)
        log.addHandler(fh)
    if stream is not None:
        ch = logging.StreamHandler(stream)
        ch.setLevel(level)
        log.addHandler(ch)
    if console:
        import sys
        cout = logging.StreamHandler(sys.stdout)
        cout.setLevel(level)
        cout.addFilter(_simple_stdout_filter())
        log.addHandler(cout)
        cerr = logging.StreamHandler(sys.stderr)
        cerr.setLevel(level)
        cerr.addFilter(_simple_stderr_filter())
        log.addHandler(cerr)
    if formatter is not None:
        for h in log.handlers:
            h.setFormatter(formatter)
    if (filename is None) and \
       (stream is None) and \
       (not console):
        log.disabled = True
    return log

def _run_command_line_solver(problem, args):
    import pybnb
    if args.nested_solver:
        problem = pybnb.futures.NestedSolver(
            problem,
            node_limit=args.nested_node_limit,
            time_limit=args.nested_time_limit,
            queue_strategy=args.nested_queue_strategy)
    else:
        nested_solver_defaults = get_default_args(
            pybnb.futures.NestedSolver.__init__)
        if args.nested_node_limit != \
           nested_solver_defaults["node_limit"]:     #pragma:nocover
            logging.getLogger("pybnb").warning(
                "The user-specified --nested-node-limit "
                "setting will be ignored. Did you forget the "
                "--nested-solver flag?")
        if args.nested_time_limit != \
           nested_solver_defaults["time_limit"]:     #pragma:nocover
            logging.getLogger("pybnb").warning(
                "The user-specified --nested-time-limit "
                "setting will be ignored. Did you forget the "
                "--nested-solver flag?")
        if args.nested_queue_strategy != \
           nested_solver_defaults["queue_strategy"]: #pragma:nocover
            logging.getLogger("pybnb").warning(
                "The user-specified --nested-queue-strategy "
                "setting will be ignored. Did you forget the "
                "--nested-solver flag?")
    solve_kwds = dict(vars(args))
    del solve_kwds["disable_mpi"]
    del solve_kwds["profile"]
    del solve_kwds["nested_solver"]
    del solve_kwds["nested_node_limit"]
    del solve_kwds["nested_time_limit"]
    del solve_kwds["nested_queue_strategy"]
    if args.disable_mpi:
        results = pybnb.solve(problem, comm=None, **solve_kwds)
    else:
        results = pybnb.solve(problem, **solve_kwds)
    return results

def create_command_line_solver(problem, parser=None):
    """Convert a given problem implementation to a
    command-line example by exposing the
    :func:`pybnb.solver.solve` function arguments using
    argparse."""
    import os
    import tempfile
    # for profiling
    try:
        import cProfile as profile
    except ImportError:                                #pragma:nocover
        import profile
    try:
        import pstats
        pstats_available=True
    except ImportError:                                #pragma:nocover
        pstats_available=False
    import pybnb
    try:
        import yaml
    except ImportError:                                #pragma:nocover
        raise ImportError("The PyYAML module is required to "
                          "run the command-line solver.")

    if parser is None:
        import argparse
        parser = argparse.ArgumentParser(
            description="Run parallel branch and bound",
            formatter_class=argparse.\
                ArgumentDefaultsHelpFormatter)

    solver_init_defaults = get_default_args(
        pybnb.Solver.__init__)
    solver_init_docs = get_keyword_docs(
        pybnb.Solver.__doc__)
    assert set(solver_init_defaults.keys()) == \
        set(solver_init_docs.keys())
    solver_init_defaults.pop("comm")
    solver_init_docs.pop("comm")
    assert len(solver_init_defaults) == len(solver_init_docs)
    for key in solver_init_defaults:
        assert solver_init_defaults[key] == \
            solver_init_docs[key]["default"]
        assert "choices" not in solver_init_docs[key]
    parser.add_argument(
        "--dispatcher-rank",
        type=int,
        default=solver_init_defaults.pop("dispatcher_rank"),
        help=solver_init_docs["dispatcher_rank"]["doc"])
    assert len(solver_init_defaults) == 0, str(solver_init_defaults)

    solve_defaults = get_default_args(
        pybnb.Solver.solve)
    solve_docs = get_keyword_docs(
        pybnb.Solver.solve.__doc__)
    solve_docs.pop("problem")
    assert set(solve_defaults.keys()) == \
        set(solve_docs.keys())
    solve_defaults.pop("best_node")
    solve_docs.pop("best_node")
    solve_defaults.pop("initialize_queue")
    solve_docs.pop("initialize_queue")
    solve_defaults.pop("scale_function")
    solve_docs.pop("scale_function")
    solve_defaults.pop("log")
    solve_docs.pop("log")
    assert len(solve_defaults) == len(solve_docs)
    for key in solve_defaults:
        assert solve_defaults[key] == \
            solve_docs[key]["default"]
        assert "choices" not in solve_docs[key]
        if key == "queue_strategy":
            solve_docs[key]["choices"] = \
                [v_.value for v_ in pybnb.QueueStrategy]
            assert "**(D)**" in solve_docs[key]["doc"]
            solve_docs[key]["doc"] = \
                ("**(D)** Sets the strategy for prioritizing "
                 "nodes in the central dispatcher queue. Can "
                 "also be set to a comma-separated list of "
                 "choices to define a lexicographic sorting "
                 "strategy.")
    class _QueueStrategyJoin(argparse.Action):    #pragma:nocover
        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            if nargs is not None:
                raise ValueError("nargs not allowed")
            super(_QueueStrategyJoin, self).__init__(option_strings,
                                                     dest,
                                                     **kwargs)
        def __call__(self, parser, namespace, values, option_string=None):
            if values in solve_docs["queue_strategy"]["choices"]:
                namespace.queue_strategy = values
            else:
                assert "," in values
                vals = tuple(v.strip() for v in values.split(',') if v.strip())
                assert len(vals) > 0
                assert all(v in solve_docs["queue_strategy"]["choices"]
                           for v in vals)
                namespace.queue_strategy = vals
    class _QueueStrategyChoices(object):          #pragma:nocover
        def __contains__(self, val):
            if val in solve_docs["queue_strategy"]["choices"]:
                return True
            if "," not in val:
                return False
            vals = [v.strip() for v in val.split(',') if v.strip()]
            if len(vals) == 0:
                return False
            return all(v in solve_docs["queue_strategy"]["choices"]
                       for v in vals)
        def __iter__(self):
            return solve_docs["queue_strategy"]["choices"].__iter__()
    parser.add_argument(
        "--best-objective",
        type=float,
        default=solve_defaults.pop("best_objective"),
        help=solve_docs["best_objective"]["doc"])
    tmp_ = solve_defaults.pop("disable_objective_call")
    assert not tmp_
    del tmp_
    parser.add_argument(
        "--disable-objective-call",
        default=False,
        action="store_true",
        help=solve_docs["disable_objective_call"]["doc"])
    parser.add_argument(
        "--queue-strategy",
        type=str,
        choices=_QueueStrategyChoices(),
        action=_QueueStrategyJoin,
        default=solve_defaults.pop("queue_strategy"),
        help=solve_docs["queue_strategy"]["doc"])
    parser.add_argument(
        "--absolute-gap",
        type=float,
        default=solve_defaults.pop("absolute_gap"),
        help=solve_docs["absolute_gap"]["doc"])
    parser.add_argument(
        "--relative-gap",
        type=float,
        default=solve_defaults.pop("relative_gap"),
        help=solve_docs["relative_gap"]["doc"])
    parser.add_argument(
        "--objective-stop",
        type=float,
        default=solve_defaults.pop("objective_stop"),
        help=solve_docs["objective_stop"]["doc"])
    parser.add_argument(
        "--bound-stop",
        type=float,
        default=solve_defaults.pop("bound_stop"),
        help=solve_docs["bound_stop"]["doc"])
    parser.add_argument(
        "--node-limit",
        type=int,
        default=solve_defaults.pop("node_limit"),
        help=solve_docs["node_limit"]["doc"])
    parser.add_argument(
        "--time-limit",
        type=float,
        default=solve_defaults.pop("time_limit"),
        help=solve_docs["time_limit"]["doc"])
    def _float_or_None(val):                      #pragma:nocover
        if val == "None":
            return None
        return float(val)
    parser.add_argument(
        "--queue-tolerance",
        type=_float_or_None,
        default=solve_defaults.pop("queue_tolerance"),
        help=solve_docs["queue_tolerance"]["doc"])
    parser.add_argument(
        "--branch-tolerance",
        type=_float_or_None,
        default=solve_defaults.pop("branch_tolerance"),
        help=solve_docs["branch_tolerance"]["doc"])
    parser.add_argument(
        "--comparison-tolerance",
        type=float,
        default=solve_defaults.pop("comparison_tolerance"),
        help=solve_docs["comparison_tolerance"]["doc"])
    parser.add_argument(
        "--log-interval-seconds",
        type=float,
        default=solve_defaults.pop("log_interval_seconds"),
        help=solve_docs["log_interval_seconds"]["doc"])
    val = solve_defaults.pop("log_new_incumbent")
    assert val
    parser.add_argument(
        "--disable-log-new-incumbent",
        action="store_false",
        dest="log_new_incumbent",
        default=True,
        help=solve_docs["log_new_incumbent"]["doc"])
    val = solve_defaults.pop("disable_signal_handlers")
    assert not val
    parser.add_argument(
        "--disable-signal-handlers",
        action="store_true",
        default=False,
        help=solve_docs["disable_signal_handlers"]["doc"])
    assert len(solve_defaults) == 0, str(solve_defaults)

    parser.add_argument(
        "--log-filename", type=str, default=None,
        help=("A filename to store solver output into."))
    parser.add_argument(
        "--results-filename", type=str, default=None,
        help=("When set, saves the solver results into a "
              "YAML-formatted file with the given name."))
    parser.add_argument(
        "--disable-mpi", default=False,
        action="store_true",
        help=("Do not attempt to import mpi4py.MPI. "
              "Enabling this option is equivalent to "
              "creating a Solver with `comm=None`."))
    if pstats_available:
        parser.add_argument(
            "--profile", dest="profile", type=int, default=0,
            help=("Enable profiling by setting this "
                  "option to a positive integer (the "
                  "maximum number of functions to "
                  "profile)."))
    parser.add_argument('--version',
                        action='version',
                        version='pybnb '+str(pybnb.__version__))
    nested_solver_defaults = get_default_args(
        pybnb.futures.NestedSolver.__init__)
    nested_solver_docs = get_keyword_docs(
        pybnb.futures.NestedSolver.__doc__)
    nested_solver_docs.pop("problem")
    assert len(nested_solver_defaults) == len(nested_solver_docs)
    parser.add_argument(
        "--nested-solver",
        action="store_true",
        default=False,
        help=("**(W)** Wraps the problem in a "
              ":class:`pybnb.futures.NestedSolver` object. "
              "See additional --nested-solver-* options."))
    parser.add_argument(
        "--nested-node-limit",
        type=int,
        default=nested_solver_defaults.pop("node_limit"),
        help=nested_solver_docs["node_limit"]["doc"])
    parser.add_argument(
        "--nested-time-limit",
        type=float,
        default=nested_solver_defaults.pop("time_limit"),
        help=nested_solver_docs["time_limit"]["doc"])
    parser.add_argument(
        "--nested-queue-strategy",
        type=str,
        choices=_QueueStrategyChoices(),
        action=_QueueStrategyJoin,
        default=nested_solver_defaults.pop("queue_strategy"),
        help=nested_solver_docs["queue_strategy"]["doc"])
    assert len(nested_solver_defaults) == 0,\
        str(nested_solver_defaults)

    args = parser.parse_args()

    try:
        import mpi4py
    except ImportError:                                #pragma:nocover
        if not args.disable_mpi:
            raise ImportError("The mpi4py module is not "
                              "available. To run this script "
                              "without it, use the '--disable-mpi' "
                              "option")

    if pstats_available and (args.profile):            #pragma:nocover
        #
        # Call the main routine with profiling.
        #
        handle, tfile = tempfile.mkstemp()
        os.close(handle)
        try:
            profile.runctx("_run_command_line_solver(problem, args)",
                           globals(),
                           locals(),
                           tfile)
            p = pstats.Stats(tfile).strip_dirs()
            p.sort_stats("time", "cumulative")
            p = p.print_stats(args.profile)
            p.print_callers(args.profile)
            p.print_callees(args.profile)
            p = p.sort_stats("cumulative","calls")
            p.print_stats(args.profile)
            p.print_callers(args.profile)
            p.print_callees(args.profile)
            p = p.sort_stats("calls")
            p.print_stats(args.profile)
            p.print_callers(args.profile)
            p.print_callees(args.profile)
        finally:
            os.remove(tfile)
    else:
        _run_command_line_solver(problem, args)



import pybnb

class UnboundedMax(pybnb.Problem):

    #
    # Implement Problem abstract methods
    #

    def sense(self): return pybnb.maximize
    def objective(self): return self.unbounded_objective()
    def bound(self): return self.unbounded_objective()
    def save_state(self, node): pass
    def load_state(self, node): pass
    def branch(self): raise NotImplementedError()




# configure a very basic logger for the module
def _configLogging():
    import logging
    logger = logging.getLogger('pybnb')
    logger.setLevel(logging.WARNING)
    formatter = logging.Formatter(
        '%(levelname)s(%(name)s): %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
_configLogging()
del _configLogging

from pybnb.__about__ import __version__
from pybnb.configuration import config
from pybnb.common import (minimize,
                          maximize,
                          inf,
                          nan,
                          QueueStrategy,
                          SolutionStatus,
                          TerminationCondition)
from pybnb.priority_queue import (CustomPriorityQueue,
                                  register_queue_type)
from pybnb.node import Node
from pybnb.problem import Problem
from pybnb.solver_results import SolverResults
from pybnb.solver import (Solver,
                          solve)



import logging

from pybnb.common import minimize
from pybnb.convergence_checker import ConvergenceChecker
from pybnb.node import Node
from pybnb.solver import Solver
from pybnb.problem import Problem
from pybnb.dispatcher import DispatcherQueueData
from pybnb.misc import get_simple_logger
from pybnb.futures import _RedirectHandler

from .common import mpi_available

from six import StringIO
from runtests.mpi import MPITest

def _get_logging_baseline(size):
    out = \
"""[WARNING] 0: warning
[ERROR] 0: error
[CRITICAL] 0: critical"""
    for i in range(1,size):
        out += ("""
[WARNING] %d: warning
[ERROR] %d: error
[CRITICAL] %d: critical""") % (i,i,i)
    return out

class DummyProblem(Problem):

    def sense(self): return minimize
    def objective(self): return 0.0
    def bound(self): return 0.0
    def save_state(self, node): pass
    def load_state(self, node): pass
    def branch(self): return ()

def _logging_redirect_check(comm):
    opt = Solver(comm=comm)
    p = DummyProblem()
    log = logging.Logger(None,
                         level=logging.WARNING)
    log.addHandler(_RedirectHandler(opt._disp))
    if opt.is_dispatcher:
        assert (comm is None) or (comm.rank == 0)
        root = Node()
        p.save_state(root)
        root.objective = p.infeasible_objective()
        root.bound = p.unbounded_objective()
        initialize_queue = DispatcherQueueData(
            nodes=[root],
            worst_terminal_bound=None,
            sense=p.sense())
        out = StringIO()
        formatter = logging.Formatter("[%(levelname)s] %(message)s")
        opt._disp.initialize(p.infeasible_objective(),
                             None,
                             initialize_queue,
                             "bound",
                             ConvergenceChecker(p.sense()),
                             None, None,
                             get_simple_logger(console=True,
                                               stream=out,
                                               level=logging.DEBUG,
                                               formatter=formatter),
                             0.0,
                             True)
        log.debug("0: debug")
        log.info("0: info")
        log.warning("0: warning")
        log.error("0: error")
        log.critical("0: critical")
        if (comm is not None) and (comm.size > 1):
            opt._disp.serve()
    else:
        assert comm is not None
        if comm.size > 1:
            for i in range(1, comm.size):
                if comm.rank == i:
                    log.debug(str(comm.rank)+": debug")
                    log.info(str(comm.rank)+": info")
                    log.warning(str(comm.rank)+": warning")
                    log.error(str(comm.rank)+": error")
                    log.critical(str(comm.rank)+": critical")
                opt.worker_comm.Barrier()
            if opt.worker_comm.rank == 0:
                opt._disp.stop_listen()
    if comm is not None:
        comm.Barrier()
    if opt.is_dispatcher:
        assert ('\n'.join(out.getvalue().splitlines()[7:])) == \
                _get_logging_baseline(comm.size if comm is not None else 1)

def test_logging_redirect_nocomm():
    _logging_redirect_check(None)

if mpi_available:

    @MPITest(commsize=[1, 2, 3])
    def test_logging_redirect(comm):
        _logging_redirect_check(comm)



import itertools

from runtests.mpi import MPITest

import pybnb

from .common import mpi_available

def left_child(i):
    return 2*i + 1

def right_child(i):
    return 2*i + 2

def log2floor(n):
    assert n > 0
    return n.bit_length() - 1

def height(size):
    return log2floor(size)

def set_none(heap, i):
    if i < len(heap):
        heap[i] = None
        set_none(heap, left_child(i))
        set_none(heap, right_child(i))

def set_one(heap, i):
    if i < len(heap):
        if heap[i] is not None:
            heap[i] = 1
            set_one(heap, left_child(i))
            set_one(heap, right_child(i))

def is_terminal(heap, i):
    N = len(heap)
    c1 = left_child(i)
    c2 = right_child(i)
    return ((c1 >= N) or (heap[c1] is None)) and \
           ((c2 >= N) or (heap[c2] is None))

def get_bound(heap, how=min):
    N = len(heap)
    assert N >= 1
    assert heap[0] is not None
    terminal_bounds = []
    for i in range(N):
        if (heap[i] is not None) and is_terminal(heap, i):
            terminal_bounds.append(heap[i])
    return how(terminal_bounds)

def powerset(iterable):
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s)+1))

def gen_heaps(k):
    for heap_size in range(1,2**(k+1)):
        h = height(heap_size)
        for level_none in range(h,h+1):
            level_nodes = list(filter(lambda x: x < heap_size,
                                      range((2**level_none)-1,(2**(level_none+1))-1)))
            for none_list in sorted(powerset(level_nodes)):
                if len(none_list) == len(level_nodes):
                    continue
                heap_master = [0]*heap_size
                for i in none_list:
                    set_none(heap_master, i)
                for level in range(0,h+1):
                    nodes = filter(lambda x: (x < heap_size) and (heap_master[x] is not None),
                                   range((2**level)-1,(2**(level+1))-1))
                    for nodes_list in sorted(powerset(nodes)):
                        if (len(nodes_list) == 0) and \
                           level != 0:
                            continue
                        heap = [0]*heap_size
                        for i in none_list:
                            set_none(heap, i)
                        for i in nodes_list:
                            set_one(heap, i)
                        yield heap

class Discrete(pybnb.Problem):

    def __init__(self,
                 sense,
                 objectives,
                 bound_bheap,
                 default_objective):
        assert len(bound_bheap) >= 1
        self._sense = sense
        self._objectives = objectives
        self._bound_bheap = bound_bheap
        self._default_objective = default_objective
        self._heap_idx = 0

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return self._sense

    def objective(self):
        return self._objectives.get(self._heap_idx,
                                    self._default_objective)

    def bound(self):
        return self._bound_bheap[self._heap_idx]

    def save_state(self, node):
        node.state = self._heap_idx

    def load_state(self, node):
        self._heap_idx = node.state

    def branch(self):
        i = self._heap_idx
        assert i >= 0
        assert i < len(self._bound_bheap)
        left_idx =  2*i + 1
        if (left_idx < len(self._bound_bheap)) and \
           (self._bound_bheap[left_idx] is not None):
            child = pybnb.Node()
            child.state = left_idx
            yield child
        right_idx = 2*i + 2
        if (right_idx < len(self._bound_bheap)) and \
           (self._bound_bheap[right_idx] is not None):
            child = pybnb.Node()
            child.state = right_idx
            yield child

class SmallHeap(pybnb.Problem):
    def __init__(self):
        self._heap_index = 0
        self._max_heap_index = 4
    def sense(self):
        return pybnb.minimize
    def objective(self):
        if self._heap_index == 0:
            return 3
        elif self._heap_index in (1,2):
            return 2
        elif self._heap_index == 3:
            return 0
        else:
            assert self._heap_index == 4
            return 1
    def bound(self):
        if self._heap_index == 0:
            return -3
        elif self._heap_index in (1,2):
            return -2
        else:
            assert self._heap_index in (3,4)
            return -1
    def save_state(self, node):
        node.state = self._heap_index
    def load_state(self, node):
        self._heap_index = node.state
    def branch(self):
        i = self._heap_index
        assert 0 <= i <= self._max_heap_index
        left_index =  2*i + 1
        if left_index <= self._max_heap_index:
            child = pybnb.Node()
            child.state = left_index
            yield child
        right_index = 2*i + 2
        if right_index <= self._max_heap_index:
            child = pybnb.Node()
            child.state = right_index
            yield child

def _test_heaps(comm):
    solver = pybnb.Solver(comm=comm)
    if comm is not None:
        if comm.rank == 0:
            pass
        elif comm.rank == 1:
            pass
        elif comm.rank == 3:
            pass

    problem = SmallHeap()
    results = solver.solve(problem,
                           queue_strategy='breadth')
    assert results.solution_status == "feasible"
    assert results.termination_condition == "no_nodes"
    assert results.objective == 0
    assert results.bound == -2
    assert results.best_node.objective == 0
    assert results.best_node.bound == -1
    assert results.best_node.state == 3
    _uuid = results.best_node._uuid
    queue = solver.save_dispatcher_queue()
    if solver.is_dispatcher:
        assert queue.bound() == -2
        assert queue.worst_terminal_bound == -2
        assert len(queue.nodes) == 0
    results = solver.solve(problem,
                           initialize_queue=queue,
                           best_node=results.best_node)
    assert results.solution_status == "feasible"
    assert results.termination_condition == "no_nodes"
    assert results.objective == 0
    assert results.bound == -2
    assert results.best_node.objective == 0
    assert results.best_node.bound == -1
    assert results.best_node.state == 3
    assert results.best_node._uuid == _uuid
    queue = solver.save_dispatcher_queue()
    if solver.is_dispatcher:
        assert queue.bound() == -2
        assert queue.worst_terminal_bound == -2
        assert len(queue.nodes) == 0

    for heap in gen_heaps(2):
        heap_bound = get_bound(heap)
        node_list = [None, len(heap)] + \
            [i for i in range(len(heap))
             if heap[i] is not None]
        # min
        for objective_node in node_list:
            if objective_node is not None:
                problem = Discrete(
                    pybnb.minimize,
                    {objective_node: 1},
                    heap,
                    default_objective=2)
            else:
                problem = Discrete(
                    pybnb.minimize,
                    {},
                    heap,
                    default_objective=1)
            results = solver.solve(problem, log=None)
            if objective_node == len(heap):
                assert results.objective == 2
            else:
                assert results.objective == 1
            assert results.bound == heap_bound
        # max
        heap_bound = -heap_bound
        heap = [-b_ if (b_ is not None) else None
                for b_ in heap]
        for objective_node in node_list:
            if objective_node is not None:
                problem = Discrete(
                    pybnb.maximize,
                    {objective_node: -1},
                    heap,
                    default_objective=-2)
            else:
                problem = Discrete(
                    pybnb.maximize,
                    {},
                    heap,
                    default_objective=-1)
            results = solver.solve(problem, log=None)
            if objective_node == len(heap):
                assert results.objective == -2
            else:
                assert results.objective == -1
            assert results.bound == heap_bound

def test_heaps_nocomm():
    _test_heaps(None)

if mpi_available:

    @MPITest(commsize=[1, 2, 4])
    def test_heaps_comm(comm):
        _test_heaps(comm)



import sys
import os
import random
import math

thisdir = os.path.dirname(os.path.abspath(__file__))
exdir = os.path.join(
    os.path.dirname(os.path.dirname(thisdir)),
    "examples")
tspdir = os.path.join(exdir,"scripts","tsp")

sys.path.insert(0,tspdir)
try:
    from tsp_naive import run_2opt
finally:
    sys.path.remove(tspdir)

class TestHeuristics(object):

    def _generate_random_matrix(self, size, symmetric=False):
        assert size > 1
        dist = [[None for j in range(size)]
               for i in range(size)]
        for i in range(size):
            dist[i][i] = 0
            for j in range(i):
                assert i != j
                dist[i][j] = random.randint(1,1000)
                if symmetric:
                    dist[j][i] = dist[i][j]
                else:
                    dist[j][i] = random.randint(1,1000)
        return dist

    def _compute_route_cost(self, dist, route):
        N = len(route)
        assert N == len(dist)+1
        assert route[0] == route[-1]
        cost = 0
        for k in range(N-1):
            assert route[k] != route[k+1]
            assert dist[route[k]][route[k+1]] != 0
            cost += dist[route[k]][route[k+1]]
        assert not math.isinf(cost)
        assert not math.isnan(cost)
        return cost

    def test_run_2opt(self):
        for n in range(1,7):
            size = 2**n
            for dist in (self._generate_random_matrix(size, symmetric=True),
                         self._generate_random_matrix(size, symmetric=False)):
                route = list(range(size))
                route.append(route[0])
                cost = self._compute_route_cost(dist, route)
                new_cost, new_route = run_2opt(dist, cost, route)
                if new_route != route:
                    assert new_cost < cost
                new_cost_ = self._compute_route_cost(dist, new_route)
                assert new_cost_ == new_cost



import os
import glob
import subprocess
import tempfile

import pytest

mpi4py_available = False
try:
    import mpi4py
    mpi4py_available = True
except ImportError:
    pass

pyomo_available = False
try:
    import pyomo.kernel as pmo
    if getattr(pmo,"version_info",(0,)*3) >= (5,4,3):  #pragma:nocover
        pyomo_available = True
except:                                                #pragma:nocover
    pass

ipopt_available = False
if pyomo_available:                                    #pragma:nocover
    from pyomo.opt.base import UnknownSolver
    ipopt = pmo.SolverFactory("ipopt")
    if ipopt is None or isinstance(ipopt, UnknownSolver):
        ipopt_available = False
    else:
        ipopt_available = \
            (ipopt.available(exception_flag=False)) and \
            ((not hasattr(ipopt,"executable")) or \
            (ipopt.executable() is not None))

yaml_available = False
try:
    import yaml
    yaml_available = True
except ImportError:
    pass

thisfile = os.path.abspath(__file__)
thisdir = os.path.dirname(thisfile)
topdir = os.path.dirname(
            os.path.dirname(thisdir))
exdir = os.path.join(topdir, "examples")
examples = []
examples.extend(glob.glob(
    os.path.join(exdir,"command_line_problems","*.py")))
examples.extend(glob.glob(
    os.path.join(exdir,"scripts","*.py")))
examples.extend(glob.glob(
    os.path.join(exdir,"scripts","tsp","tsp_naive.py")))
baselinedir = os.path.join(thisdir, "example_baselines")

assert os.path.exists(exdir)
assert thisfile not in examples

tdict = {}
for fname in examples:
    basename = os.path.basename(fname)
    assert basename.endswith(".py")
    assert len(basename) >= 3
    basename = basename[:-3]
    if basename == "tsp_naive":
        for datafile in ('p01_d',
                         'p01_d_inf'):
            tname = "test_"+basename+"_"+datafile
            bname = os.path.join(baselinedir,
                                 basename+"_"+datafile+".yaml")
            tdict[tname] = (fname,
                            bname,
                            [os.path.join(exdir,"scripts","tsp",
                                          datafile+".txt")])
    else:
        tname = "test_"+basename
        bname = os.path.join(baselinedir,basename+".yaml")
        tdict[tname] = (fname,bname,None)
assert len(tdict) == len(examples) + 1

assert "test_binary_knapsack" in tdict
assert len(tdict["test_binary_knapsack"]) == 3
assert "test_binary_knapsack_nested" not in tdict
tdict["test_binary_knapsack_nested"] = \
    (tdict["test_binary_knapsack"][0],
     tdict["test_binary_knapsack"][1],
     ["--nested-solver",
      "--nested-node-limit=100",
      "--nested-queue-strategy=random"])

scenarios = []
for p in [1,2,4]:
    for name in sorted(tdict):
        scenarios.append((name, p))

@pytest.mark.parametrize(("example_name", "procs"),
                         scenarios)
@pytest.mark.example
def test_example(example_name, procs):
    if not yaml_available:
        pytest.skip("yaml is not available")
    if example_name in ("test_bin_packing",
                        "test_rosenbrock_2d",
                        "test_range_reduction_pyomo"):
        if not (pyomo_available and ipopt_available):
            pytest.skip("Pyomo or Ipopt is not available")
    if (example_name == "test_simple") or \
       ("tsp_naive" in example_name):
        if not mpi4py_available:
            pytest.skip("MPI is not available")
    if (not mpi4py_available) and (procs > 1):
        pytest.skip("MPI is not available")
    filename, baseline_filename, options = tdict[example_name]
    cmd = ["python", filename]
    if options is None:
        options = []
    assert os.path.exists(filename)
    fid, results_filename = tempfile.mkstemp()
    os.close(fid)
    try:
        if procs == 1:
            if ("range_reduction_pyomo" in example_name) or \
               ("tsp_naive" in example_name):
                rc = subprocess.call(cmd + \
                                     ["--results-file",
                                      results_filename] + \
                                     options)
            elif example_name == "test_simple":
                rc = subprocess.call(cmd + options)
            else:
                rc = subprocess.call(cmd + \
                                     ["--disable-mpi",
                                      "--results-file",
                                      results_filename] + \
                                     options)
        else:
            assert procs > 1
            if subprocess.call(["mpirun",
                                "--allow-run-as-root",
                                "--version"],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT):
                rc = subprocess.call(["mpirun","-np", str(procs)] + \
                                     cmd + \
                                     ["--results-file",
                                      results_filename] + \
                                     options)
            else:
                rc = subprocess.call(["mpirun", "--allow-run-as-root",
                                      "-np", str(procs)] + \
                                     cmd + \
                                     ["--results-file",
                                      results_filename] + \
                                     options)
        assert rc == 0
        if example_name == "test_simple":
            assert not os.path.exists(baseline_filename)
            return
        with open(results_filename) as f:
            results = yaml.load(f)
        with open(baseline_filename) as f:
            baseline_results = yaml.load(f)
        assert len(baseline_results) < len(results)
        for key in baseline_results:
            if type(baseline_results[key]) is float:
                assert round(baseline_results[key], 4) == \
                    round(results[key], 4)
            else:
                assert baseline_results[key] == results[key]
    finally:
        os.remove(results_filename)



"""
Configuration settings for node serialization.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import os
import platform
import pickle

from pybnb import __version__

class Configuration(object):
    """The main configuration object.

    Attributes
    ----------
    SERIALIZER : str, {'pickle', 'dill'}
        The name of serialization module used to transmit
        node state. (default: "pickle")
    SERIALIZER_PROTOCOL_VERSION : int
        The protocol argument passed to the `dumps` function
        of the selected serialization module.
        (default: pickle.HIGHEST_PROTOCOL)
    MARSHAL_PROTOCOL_VERSION : int
        The version argument passed to the
        :func:`marshal.dumps` function. (default: 2)
    """
    __slots__ = ("SERIALIZER",
                 "SERIALIZER_PROTOCOL_VERSION",
                 "MARSHAL_PROTOCOL_VERSION")

    def __init__(self):
        self.reset()

    def reset(self, use_environment=True):
        """Reset the configuration to default settings.

        Parameters
        ----------
        use_environment : bool, optional
            Controls whether or not to check for environment
            variables to overwrite the default
            settings. (default: True)
        """
        self.SERIALIZER = "pickle"
        self.SERIALIZER_PROTOCOL_VERSION = pickle.HIGHEST_PROTOCOL
        self.MARSHAL_PROTOCOL_VERSION = 2
        if use_environment:
            # process environment variables
            PREFIX = "PYBNB_"
            for symbol in self.__slots__:
                if PREFIX+symbol in os.environ:
                    default = getattr(self, symbol)
                    envvalue = type(default)(os.environ[PREFIX+symbol])
                    setattr(self, symbol, envvalue)

    def __str__(self):
        out =  "pybnb version: %s\n" % __version__
        out += ("loaded from: %s\n"
                % (os.path.dirname(__file__)))
        out += ("python version: %s %s (%s, %s)\n"
                % (platform.python_implementation(),
                   platform.python_version(),
                   platform.system(),
                   os.name))
        out += "configuration:"
        for key in self.__slots__:
            out += ("\n - %s: %s" % (key,
                                     getattr(self, key)))
        return out

config = Configuration()

if __name__ == "__main__":                        #pragma:nocover
    print(config)



"""
Branch-and-bound node implementation.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import uuid

from pybnb.configuration import config

import six

if not six.PY2:
    import pickle
else:
    import cPickle as pickle

_serializer_modules = {}
_serializer_modules["pickle"] = pickle
_serializer_modules["dill"] = None

def _get_dill():
    assert config.SERIALIZER == "dill"
    import dill
    _serializer_modules["dill"] = dill
    return dill

def dumps(obj):
    """Return the serialized representation of the object as
    a bytes object, using the serialization module set in
    the current configuration."""
    try:
        mod = _serializer_modules[config.SERIALIZER]
    except KeyError:
        raise ValueError("Invalid serializer '%s'. "
                         "Valid choices are: ['pickle', 'dill']"
                         % (config.SERIALIZER))
    if mod is None:
        mod = _get_dill()
    return mod.dumps(
        obj,
        protocol=config.SERIALIZER_PROTOCOL_VERSION)

def loads(obj):
    """Read and return an object from the given serialized
    data, using the serialization module set in the current
    configuration."""
    try:
        mod = _serializer_modules[config.SERIALIZER]
    except KeyError:
        raise ValueError("Invalid serializer '%s'. "
                         "Valid choices are: ['pickle', 'dill']"
                         % (config.SERIALIZER))
    if mod is None:
        mod = _get_dill()
    return mod.loads(obj)

class _SerializedNode(object):
    """A helper object used by the distributed dispatcher
    for lightweight handling of serialized nodes."""
    __slots__ = ("objective",
                 "bound",
                 "tree_depth",
                 "queue_priority",
                 "_uuid",
                 "data")
    def __init__(self, slots):
        (self.objective,
         self.bound,
         self.tree_depth,
         self.queue_priority,
         self._uuid,
         self.data) = slots

    def _generate_uuid(self):
        self._uuid = uuid.uuid4().hex

    @property
    def slots(self):
        return (self.objective,
                self.bound,
                self.tree_depth,
                self.queue_priority,
                self._uuid,
                self.data)

    @staticmethod
    def to_slots(node):
        return (node.objective,
                node.bound,
                node.tree_depth,
                node.queue_priority,
                node._uuid,
                dumps(node.state))

    @classmethod
    def from_node(cls, node):
        return cls(cls.to_slots(node))

    @staticmethod
    def restore_node(slots):
        node = Node()
        (node.objective,
         node.bound,
         node.tree_depth,
         node.queue_priority,
         node._uuid,
         node.state) = slots
        node.state = loads(node.state)
        return node

class Node(object):
    """A branch-and-bound node that stores problem state.

    Attributes
    ----------
    objective : float
        The objective value for the node.
    bound : float
        The bound value for the node.
    tree_depth : int
        The tree depth of the node (0-based).
    queue_priority : float or tuple of floats
        The queue priority of the node.
    state
        The user specified node state.
    """
    __slots__ = ("objective",
                 "bound",
                 "tree_depth",
                 "queue_priority",
                 "_uuid",
                 "state")

    def __init__(self):
        self.objective = None
        self.bound = None
        self.tree_depth = None
        self.queue_priority = None
        self._uuid = None
        self.state = None

    def _generate_uuid(self):
        self._uuid = uuid.uuid4().hex

    def resize(self, *args, **kwds):
        raise NotImplementedError(
            "It is no longer necessary to call "
            "node.resize(...). Simply assign any object "
            "to the node.state attribute. It no longer "
            "needs to be a numpy array. The node state "
            "must be serialize-able using pickle or dill "
            "in order to be compatible with the MPI-based "
            "parallel solver.")

    def __str__(self):
        out = \
            ("Node(objective=%s,\n"
             "     bound=%s,\n"
             "     tree_depth=%s)"
             % (self.objective,
                self.bound,
                self.tree_depth))
        return out

    def new_child(self):
        child = Node()
        child.objective = self.objective
        child.bound = self.bound
        child.tree_depth = self.tree_depth + 1
        assert child.queue_priority is None
        assert child.state is None
        return child



"""
Branch-and-bound solver implementation.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""

import sys
import time
import array
import math

from pybnb.common import (minimize,
                          maximize,
                          QueueStrategy,
                          TerminationCondition,
                          SolutionStatus)
from pybnb.problem import (_SolveInfo,
                           _SimpleSolveInfoCollector,
                           _ProblemWithSolveInfoCollection)
from pybnb.misc import (MPI_InterruptHandler,
                        time_format,
                        as_stream,
                        get_simple_logger,
                        get_default_args)
from pybnb.node import Node
from pybnb.solver_results import SolverResults
from pybnb.convergence_checker import (_default_scale,
                                       ConvergenceChecker)
from pybnb.dispatcher_proxy import DispatcherProxy
from pybnb.dispatcher import (DispatcherLocal,
                              DispatcherDistributed,
                              DispatcherQueueData)
from pybnb.futures import NestedSolver

try:
    import mpi4py
except ImportError:                               #pragma:nocover
    pass

import six

class _notset(object):
    pass

# this is defined at the bottom of this file
_solve_defaults = None

class Solver(object):
    """A branch-and-bound solver.

    Parameters
    ----------
    comm : ``mpi4py.MPI.Comm``, optional
        The MPI communicator to use. If unset, the
        mpi4py.MPI.COMM_WORLD communicator will be
        used. Setting this keyword to None will disable the
        use of MPI and avoid an attempted import of
        mpi4py.MPI (which avoids triggering a call to
        `MPI_Init()`).
    dispatcher_rank : int, optional
        The process with this rank will be designated as the
        dispatcher process. If MPI functionality is disabled
        (by setting comm=None), this keyword must be 0.
        (default: 0)
    """

    def _check_for_old_branch_signature(self, problem):
        if not self.is_dispatcher:
            return
        import inspect
        argname = None
        if not six.PY2:
            # py3 does not include selse argument of class methods
            sig = inspect.signature(problem.branch)
            if len(sig.parameters) == 1:
                argname = list(sig.parameters.keys())[0]
        else:
            # py2 includes self argument of class methods
            sig = inspect.getargspec(problem.branch)
            if len(sig.args) == 2:
                argname = sig.args[1]
        if argname is not None:
            raise TypeError("The pybnb solver has detected that "
                            "the 'branch' method on your problem "
                            "uses the old call signature compatible "
                            "with pybnb 0.4.0 and earlier. To make "
                            "your problem compatible with the most "
                            "recent version of pybnb, please remove "
                            "the '%s' argument from this method, and "
                            "replace '%s.new_child()' with "
                            "'pybnb.Node()'." % (argname, argname))

    def __init__(self,
                 comm=_notset,
                 dispatcher_rank=0):
        mpi = True
        if comm is None:
            mpi = False
        self._comm = None
        self._worker_flag = None
        self._dispatcher_flag = None
        self._disp = None
        self._time = None
        if mpi:
            import mpi4py.MPI
            assert mpi4py.MPI.Is_initialized()
            assert comm is not None
            if comm is _notset:
                comm = mpi4py.MPI.COMM_WORLD
            if (int(dispatcher_rank) != dispatcher_rank) or \
               (dispatcher_rank < 0) or \
               (dispatcher_rank >= comm.size):
                raise ValueError("The 'dispatcher_rank' keyword "
                                 "has been set to %s, which is not "
                                 "an available rank given the "
                                 "size of the MPI communicator (%d)."
                                 % (dispatcher_rank, comm.size))
            self._comm = comm
            if comm.size > 1:
                dispatcher_rank = int(dispatcher_rank)
                if comm.rank == dispatcher_rank:
                    self._disp = DispatcherDistributed(comm)
                    self._worker_flag = False
                    self._dispatcher_flag = True
                else:
                    self._disp = DispatcherProxy(comm)
                    self._worker_flag = True
                    self._dispatcher_flag = False
            else:
                self._disp = DispatcherLocal()
                self._worker_flag = True
                self._dispatcher_flag = True
            self._time = mpi4py.MPI.Wtime
        else:
            if dispatcher_rank != 0:
                raise ValueError(
                    "MPI functionality has been disabled but "
                    "the 'dispatcher_rank' keyword is set to "
                    "something other than 0.")
            assert self._comm is None
            self._disp = DispatcherLocal()
            self._worker_flag = True
            self._dispatcher_flag = True
            self._time = time.time
        assert self._worker_flag in (True, False)
        assert self._dispatcher_flag in (True, False)
        assert self._disp is not None
        assert self._time is not None
        self._solve_start = None
        self._wall_time = 0.0
        self._best_objective = None
        self._best_node = None
        self._best_node_updated = False
        self._local_solve_info = _SolveInfo()
        self._global_solve_info = None

    def _reset_local_solve_stats(self):
        self._solve_start = None
        self._wall_time = 0.0
        self._best_objective = None
        self._best_node = None
        self._best_node_updated = False
        self._local_solve_info.reset()
        self._global_solve_info = None

    def _check_update_best_node(self,
                                convergence_checker,
                                node):
        objective = node.objective
        assert objective is not None
        assert not math.isnan(objective)
        updated = False
        if (objective != convergence_checker.infeasible_objective) and \
           ((self._best_node is None) or \
            convergence_checker.objective_improved(
                objective,
                self._best_node.objective)):
            if node._uuid is None:
                node._generate_uuid()
            self._best_node = node
            self._best_node_updated = True
            updated = True
        return updated

    def _fill_results(self, results, convergence_checker):
        infeasible_objective = \
            convergence_checker.infeasible_objective
        unbounded_objective = \
            convergence_checker.unbounded_objective
        if results.bound == infeasible_objective:
            if results.objective == infeasible_objective:
                results.solution_status = SolutionStatus.infeasible
            else:
                results.solution_status = SolutionStatus.invalid
        elif results.objective == infeasible_objective:
            results.solution_status = SolutionStatus.unknown
        elif results.objective == unbounded_objective:
            assert results.bound == unbounded_objective
            results.solution_status = SolutionStatus.unbounded
        else:
            results.absolute_gap = convergence_checker.\
                compute_absolute_gap(
                    results.bound,
                    results.objective)
            results.relative_gap = convergence_checker.\
                compute_relative_gap(
                    results.bound,
                    results.objective)
            if convergence_checker.objective_is_optimal(
                    results.objective,
                    results.bound):
                results.solution_status = SolutionStatus.invalid
                if (convergence_checker.sense == minimize) and \
                   (results.bound <= results.objective):
                    results.solution_status = SolutionStatus.optimal
                elif (convergence_checker.sense == maximize) and \
                     (results.bound >= results.objective):
                    results.solution_status = SolutionStatus.optimal
            else:
                results.solution_status = SolutionStatus.feasible

    def _solve(self,
               problem,
               best_objective,
               best_node,
               disable_objective_call,
               convergence_checker):
        is_nested_solver = False
        if isinstance(problem, NestedSolver):
            is_nested_solver = True
        if not isinstance(problem, _ProblemWithSolveInfoCollection):
            problem = _SimpleSolveInfoCollector(problem)
            problem.set_clock(self._time)
        problem.set_solve_info_object(self._local_solve_info)

        assert best_objective is not None
        self._best_objective = best_objective
        self._best_node = best_node
        infeasible_objective = problem.infeasible_objective()
        assert infeasible_objective == \
            convergence_checker.infeasible_objective
        unbounded_objective = problem.unbounded_objective()
        assert unbounded_objective == \
            convergence_checker.unbounded_objective
        first_update = True
        terminal_bound = None
        children = []
        while (1):
            update_start = self._time()
            new_best_objective = None
            if first_update:
                new_best_objective = self._best_objective
            new_best_node = None
            if self._best_node_updated:
                new_best_node = self._best_node
            self._best_node_updated = False
            (stop,
             new_best_objective,
             new_best_node,
             working_node) = self._disp.update(
                 new_best_objective,
                 new_best_node,
                 terminal_bound,
                 self._local_solve_info,
                 children)
            update_stop = self._time()

            if first_update and is_nested_solver:
                problem._initialize(self._disp,
                                    new_best_objective,
                                    disable_objective_call)

            old_best_node = self._best_node
            self._best_objective = new_best_objective
            assert (old_best_node is None) or \
                (old_best_node._uuid is not None)
            assert (new_best_node is None) or \
                (new_best_node._uuid is not None)
            updated = False
            if (new_best_node is not None) and \
               ((old_best_node is None) or \
                (new_best_node._uuid != old_best_node._uuid) or \
                first_update):
                self._best_node = new_best_node
                problem.notify_new_best_node(node=self._best_node,
                                             current=False)
            del old_best_node
            del new_best_objective
            del new_best_node
            first_update = False

            if stop:
                # make sure all processes have the exact same best
                # objective value (not just subject to tolerances)
                break
            if not is_nested_solver:
                self._local_solve_info.\
                    _increment_explored_nodes_stat(1)
                self._local_solve_info.\
                    _increment_queue_stat(
                        update_stop-update_start, 1)


            # we should not be receiving a node that
            # does not satisfy these assertions
            assert convergence_checker.eligible_for_queue(
                working_node.bound,
                self._best_objective)
            assert working_node.tree_depth >= 0

            problem.load_state(working_node)
            if is_nested_solver:
                problem._solve()
                terminal_bound = problem._queue.worst_terminal_bound
                children = problem._queue.nodes
                if problem._results.best_node is not None:
                    self._check_update_best_node(
                        convergence_checker,
                        problem._results.best_node)
                continue
            new_bound = problem.bound()
            if convergence_checker.bound_worsened(new_bound,
                                                  working_node.bound):    #pragma:nocover
                self._disp.log_warning(
                    "WARNING: Bound became worse "
                    "(old=%r, new=%r)"
                    % (working_node.bound, new_bound))
            working_node.bound = new_bound
            children = []
            if convergence_checker.eligible_for_queue(
                    working_node.bound,
                    self._best_objective):
                if not disable_objective_call:
                    working_node.objective = problem.objective()
                if convergence_checker.best_bound(
                        working_node.bound,
                        working_node.objective) != working_node.objective: #pragma:nocover
                    self._disp.log_warning(
                        "WARNING: Local node bound is worse "
                        "than local node objective (bound=%r, "
                        "objective=%r)" % (working_node.bound,
                                           working_node.objective))
                updated = self._check_update_best_node(
                    convergence_checker,
                    working_node)
                if updated:
                    problem.notify_new_best_node(node=self._best_node,
                                                 current=True)
                if (working_node.objective != unbounded_objective) and \
                    convergence_checker.eligible_for_queue(
                        working_node.bound,
                        self._best_objective) and \
                    convergence_checker.eligible_to_branch(
                        working_node.bound,
                        working_node.objective):
                    clist = problem.branch()
                    for child in clist:
                        children.append(child)
                        assert child.tree_depth is None
                        child.tree_depth = working_node.tree_depth + 1
                        if child.bound is None:
                            child.bound = working_node.bound
                        elif convergence_checker.bound_worsened(
                                child.bound,
                                working_node.bound):    #pragma:nocover
                            self._disp.log_warning(
                                "WARNING: Bound on child node "
                                "returned from branch method "
                                "is worse than parent node "
                                "(child=%r, parent=%r)"
                                % (child.bound,
                                   working_node.bound))
                        if child.objective is None:
                            child.objective = working_node.objective
            if len(children) > 0:
                terminal_bound = None
            else:
                terminal_bound = working_node.bound

        assert len(working_node) == 3
        global_bound = working_node[0]
        termination_condition = working_node[1]
        global_solve_info = working_node[2]
        return (self._best_objective,
                self._best_node,
                global_bound,
                termination_condition,
                global_solve_info)

    #
    # Interface
    #

    @property
    def is_worker(self):
        """Indicates if this process has been designated as
        a worker."""
        return self._worker_flag

    @property
    def is_dispatcher(self):
        """Indicates if this process has been designated as
        the dispatcher."""
        return self._dispatcher_flag

    @property
    def comm(self):
        """The full MPI communicator that includes the
        dispatcher and all workers. Will be None if MPI
        functionality has been disabled."""
        return self._comm

    @property
    def worker_comm(self):
        """The worker MPI communicator. Will be None on any
        processes for which :attr:`Solver.is_worker` is
        False, or if MPI functionality has been disabled."""
        if (self._comm is None) or \
           (self._comm.size == 1):
            return self._comm
        elif not self.is_dispatcher:
            return self._disp.worker_comm
        return None

    @property
    def worker_count(self):
        """The number of worker processes associated with this solver."""
        if (self._comm is None) or \
           (self._comm.size == 1):
            return 1
        elif not self.is_dispatcher:
            return self._disp.worker_comm.size
        else:
            return len(self._disp.worker_ranks)

    def collect_worker_statistics(self):
        """Collect individual worker statistics about the
        most recent solve.

        Returns
        -------
        dict
            A dictionary whose keys are the different
            statistics collected, where each entry is a list
            storing a value for each worker.
        """
        stats = {}
        if (self.comm is not None) and \
           (self.comm.size > 1):
            num_stats = 12
            gathered = array.array('d',[0]) * (self.worker_count*num_stats)
            if self.is_worker:
                assert self.worker_comm is not None
                assert not self.is_dispatcher
                solve_info = self._local_solve_info
                mine = array.array('d',
                    [self._wall_time,
                     solve_info.total_queue_time,
                     solve_info.queue_call_count,
                     solve_info.total_objective_time,
                     solve_info.objective_call_count,
                     solve_info.total_bound_time,
                     solve_info.bound_call_count,
                     solve_info.total_branch_time,
                     solve_info.branch_call_count,
                     solve_info.total_load_state_time,
                     solve_info.load_state_call_count,
                     solve_info.explored_nodes_count])
                assert len(mine) == num_stats
                assert len(mine) == len(gathered)//self.worker_count
                self.worker_comm.Allgather([mine, mpi4py.MPI.DOUBLE],
                                           [gathered, mpi4py.MPI.DOUBLE])
                if self.worker_comm.rank == 0:
                    self.comm.Send([gathered, mpi4py.MPI.DOUBLE],
                                   self._disp.dispatcher_rank,
                                   tag=11112111)
            else:
                assert self.worker_comm is None
                assert self.is_dispatcher
                self.comm.Recv([gathered, mpi4py.MPI.DOUBLE],
                               tag=11112111)
            for i, key in enumerate(('wall_time',
                                     'queue_time',
                                     'queue_call_count',
                                     'objective_time',
                                     'objective_call_count',
                                     'bound_time',
                                     'bound_call_count',
                                     'branch_time',
                                     'branch_call_count',
                                     'load_state_time',
                                     'load_state_call_count',
                                     'explored_nodes_count')):
                items = []
                for k in range(self.worker_count):
                    items.append(gathered[k*num_stats + i])
                stats[key] = items
        else:
            assert self.is_worker
            assert self.is_dispatcher
            solve_info = self._local_solve_info
            stats['wall_time'] = [self._wall_time]
            stats['queue_time'] = [solve_info.total_queue_time]
            stats['queue_call_count'] = [solve_info.queue_call_count]
            stats['objective_time'] = \
                [solve_info.total_objective_time]
            stats['objective_call_count'] = \
                [solve_info.objective_call_count]
            stats['bound_time'] = \
                [solve_info.total_bound_time]
            stats['bound_call_count'] = \
                [solve_info.bound_call_count]
            stats['branch_time'] = \
                [solve_info.total_branch_time]
            stats['branch_call_count'] = \
                [solve_info.branch_call_count]
            stats['load_state_time'] = \
                [solve_info.total_load_state_time]
            stats['load_state_call_count'] = \
                [solve_info.load_state_call_count]
            stats['explored_nodes_count'] = \
                [solve_info.explored_nodes_count]

        return stats

    def save_dispatcher_queue(self):
        """Saves the dispatcher queue.

        Returns
        -------
        queue : :class:`pybnb.dispatcher.DispatcherQueueData` or None
            If this process is the dispatcher, this method
            will return an object storing any nodes
            currently in the dispatcher queue.  If this
            process is not the dispatcher, this method will
            return None.  The returned object can be used to
            reinitialize a solve (e.g., with different
            algorithms settings) by assigning it to the
            initialize_queue keyword of the
            :func:`Solver.solve` method.
        """
        ret = None
        if self.is_dispatcher:
            ret = self._disp.save_dispatcher_queue()
        return ret

    def solve(self,
              problem,
              best_objective=None,
              best_node=None,
              disable_objective_call=False,
              absolute_gap=1e-8,
              relative_gap=1e-4,
              scale_function=_default_scale,
              queue_tolerance=0,
              branch_tolerance=0,
              comparison_tolerance=0,
              objective_stop=None,
              bound_stop=None,
              node_limit=None,
              time_limit=None,
              initialize_queue=None,
              queue_strategy="bound",
              log_interval_seconds=1.0,
              log_new_incumbent=True,
              log=_notset,
              disable_signal_handlers=False):
        """Solve a problem using branch-and-bound.

        Note
        ----
        Parameters for this function are treated differently
        depending on whether the process is a worker or
        dispatcher. For the serial case (no MPI), the single
        process is both a worker and a dispatcher. For the
        parallel case, exactly one process is a dispatcher
        and all other processes are workers. A **(W)** in
        the parameter description indicates that it is only
        used by worker processes (ignored otherwise). A
        **(D)** in the parameter description indicates that
        it is only used by the dispatcher process (ignored
        otherwise). An **(A)** indicates that it is used by
        all processes, and it is assumed the same value is
        provided for each process; otherwise, the behavior
        is undefined.

        Parameters
        ----------
        problem : :class:`pybnb.Problem <pybnb.problem.Problem>`
            An object defining a branch-and-bound problem.
        best_objective : float, optional
            Initializes the solve with an assumed best
            objective. Both this and the best_node option
            can be set to different values on all
            processes. The dispatcher will collect all
            values and use the best. Note that setting this
            option at, or too close to, the true optimal
            objective value may prevent the solver from
            collecting a node that stores the optimal user
            state information, so use this option with
            care. The recommended way to re-continue a solve
            from a known candidate solution is to assign the
            best_node attribute of a results object to the
            best_node solve option. Also note that the best
            node will be tracked separately from the given
            initial best objective until a node is found
            that improves upon the best objective. If this
            never happens, the best_node attribute on the
            solver results may be None or may have an
            objective that is worse than the objective
            attribute of the solver results. (default: None)
        best_node : :class:`Node <pybnb.node.Node>`, optional
            Initializes the solve with an assumed best node.
            This option can (and should) be used in place of
            the best_objective option when a best node from
            a previous solve has been collected. It can also
            be assigned a node object that was created
            manually by the user. The objective attribute is
            the only property of the node that will affect
            the solve. It must be set to a numeric
            value. (default: None)
        disable_objective_call : bool, optional
            **(W)** Disables requests for an objective value from
            subproblems. (default: False)
        absolute_gap : float, optional
            **(A)** The maximum absolute difference between
            the global bound and best objective for the
            problem to be considered solved to
            optimality. Setting to `None` will disable this
            optimality check. (default: 1e-8)
        relative_gap : float, optional
            **(A)** The maximum relative difference (absolute
            difference scaled by `max{1.0,|objective|}`)
            between the global bound and best objective for
            the problem to be considered solved to
            optimality. Setting to `None` will disable this
            optimality check. (default: 1e-4)
        scale_function : function, optional
            **(A)** A function with signature `f(bound,
            objective) -> float` that returns a positive
            scale factor used to convert the absolute
            difference between the bound and objective into
            a relative difference. The relative difference
            is compared with the `relative_gap` convergence
            tolerance to determine if the solver should
            terminate. The default is equivalent to
            `max{1.0,|objective|}`.  Other examples one
            could use are `max{|bound|,|objective|}`,
            `(|bound|+|objective|)/2`, etc.
        queue_tolerance : float, optional
            **(A)** The absolute tolerance used when
            deciding if a node is eligible to enter the
            queue. The difference between the node bound and
            the incumbent objective must be greater than
            this value. The default setting of zero means
            that nodes whose bound is equal to the incumbent
            objective are not eligible to enter the
            queue. Setting this to larger values can be used
            to control the queue size, but it should be kept
            small enough to allow absolute and relative
            optimality tolerances to be met. This option can
            also be set to `None` to allow nodes with a
            bound equal to (but not greater than) the
            incumbent objective to enter the queue.
            (default: 0)
        branch_tolerance : float, optional
            **(A)** The absolute tolerance used when
            deciding if the computed objective and bound for
            a node are sufficiently different to branch into
            the node. The default value of zero means that
            branching will occur if the bound is not exactly
            equal to the objective. This option can be set
            to `None` to enable branching for nodes with a
            bound and objective that are exactly
            equal. (default: 0)
        comparison_tolerance : float, optional
            **(A)** The absolute tolerance used when
            deciding if two objective or bound values are
            sufficiently different to be considered improved
            or worsened. This tolerance controls when the
            solver considers a new incumbent objective to be
            found. It also controls when warnings are output
            about bounds becoming worse on child
            nodes. Setting this to larger values can be used
            to avoid the above solver actions due to
            insignificant numerical differences, but it is
            better to deal with these numerical issues by
            rounding numbers to a reliable precision before
            returning them from the problem methods.
            (default: 0)
        objective_stop : float, optional
            **(A)** If provided, the solve will terminate
            when a feasible objective is found that is at
            least as good as the specified value, and the
            termination_condition flag on the results object
            will be set to 'objective_limit'. If this value
            is infinite, the solve will terminate as soon as
            a finite objective is found. (default: None)
        bound_stop : float, optional
            **(A)** If provided, the solve will terminate
            when the global bound on the objective is at
            least as good as the specified value, and the
            termination_condition flag on the results object
            will be set to 'objective_limit'. If this value
            is infinite, the solve will terminate as soon as
            a finite bound is found. (default: None)
        node_limit : int, optional
            **(D)** If provided, the solve will begin to
            terminate once this many nodes have been served
            from the dispatcher queue, and the
            termination_condition flag on the results object
            will be set to 'node_limit'. (default: None)
        time_limit : float, optional
            **(D)** If provided, the solve will begin to
            terminate once this amount of time has passed,
            and the termination_condition flag on the
            results object will be set to 'time_limit'. Note
            that the solve may run for an arbitrarily longer
            amount of time, depending how long worker
            processes spend completing their final
            task. (default: None)
        initialize_queue : :class:`pybnb.dispatcher.DispatcherQueueData`, optional
            **(D)** Initializes the dispatcher queue with
            that remaining from a previous solve (obtained
            by calling :func:`Solver.save_dispatcher_queue`
            after the solve). If left as None, the queue
            will be initialized with a single root node
            created by calling :func:`problem.save_state
            <pybnb.problem.Problem.save_state>`.
            (default: None)
        queue_strategy : :class:`QueueStrategy <pybnb.common.QueueStrategy>` or tuple
            **(D)** Sets the strategy for prioritizing nodes
            in the central dispatcher queue. See the
            :class:`QueueStrategy
            <pybnb.common.QueueStrategy>` enum for the list
            of acceptable values. This keyword can be
            assigned one of the enumeration attributes or an
            equivalent string name. This keyword can also be
            assigned a tuple of choices to define a
            lexicographic sorting strategy.
            (default: 'bound')
        log_interval_seconds : float, optional
            **(D)** The approximate time (in seconds)
            between solver log updates. More time may pass
            between log updates if no updates have been
            received from worker processes, and less time
            may pass if a new incumbent objective is
            found. (default: 1.0)
        log_new_incumbent : bool, optional
            **(D)** Controls whether updates to the best
            objective are logged immediately (overriding the
            log interval). Setting this to false can be
            useful when frequent updates to the incumbent
            are expected and the additional logging slows
            down the dispatcher. (default: True)
        log : logging.Logger, optional
            **(D)** A log object where solver output should
            be sent. The default value causes all output to
            be streamed to the console. Setting to None
            disables all output.
        disable_signal_handlers : bool, optional
            **(D)** Indicates whether or disable the
            registering of signal handlers that allow
            gracefully terminating a solve early.
            (default: False)

        Returns
        -------
        results : :class:`SolverResults <pybnb.solver_results.SolverResults>`
            An object storing information about the solve.
        """
        self._check_for_old_branch_signature(problem)
        self._reset_local_solve_stats()
        self._solve_start = self._time()

        if best_objective is None:
            best_objective = problem.infeasible_objective()
        assert not math.isnan(best_objective)

        if best_node is not None:
            if (best_node.objective is None) or \
               math.isnan(best_node.objective):
                raise ValueError("The best_node objective "
                                 "attribute must be set to "
                                 "a numeric value.")
            if best_node._uuid is None:
                best_node._generate_uuid()

        results = SolverResults()
        convergence_checker = ConvergenceChecker(
            problem.sense(),
            absolute_gap=absolute_gap,
            relative_gap=relative_gap,
            scale_function=scale_function,
            queue_tolerance=queue_tolerance,
            branch_tolerance=branch_tolerance,
            comparison_tolerance=comparison_tolerance,
            objective_stop=objective_stop,
            bound_stop=bound_stop)
        problem.notify_solve_begins(self.comm,
                                    self.worker_comm,
                                    convergence_checker)

        orig = Node()
        problem.save_state(orig)
        try:
            if self.is_dispatcher:
                if log is _notset:
                    log = get_simple_logger()
                if not isinstance(queue_strategy,
                                  (six.string_types,
                                   QueueStrategy)):
                    queue_strategy = tuple(qs.value \
                        if isinstance(qs, QueueStrategy) else qs
                        for qs in queue_strategy)
                elif isinstance(queue_strategy,
                                QueueStrategy):
                    queue_strategy = \
                        queue_strategy.value
                if log is not None:
                    changed = False
                    locals_ = locals()
                    for key_ in sorted(_solve_defaults):
                        if key_ == 'log':
                            continue
                        default_ = _solve_defaults[key_]
                        val_ = locals_[key_]
                        if key_ == 'best_objective':
                            default_ = problem.infeasible_objective()
                        if val_ != default_:
                            if not changed:
                                log.info('\nUsing non-default solver options:')
                            changed = True
                            if key_ == 'initialize_queue':
                                default_ = "<root>"
                                val_ = "Queue(size=%s)" % (len(val_.nodes))
                            elif key_ == 'best_node':
                                val_ = "Node(objective=%.7g)" % (val_.objective)
                            log.info(' - %s: %s (default: %s)'
                                     % (key_, val_, default_))
                    if changed:
                        log.info('')
                if initialize_queue is None:
                    root = Node()
                    root.tree_depth = 0
                    root.queue_priority = 0
                    root.bound = problem.unbounded_objective()
                    root.objective = problem.infeasible_objective()
                    root.state = orig.state
                    initialize_queue = DispatcherQueueData(
                        nodes=[root],
                        worst_terminal_bound=None,
                        sense=convergence_checker.sense)
                self._disp.initialize(
                    best_objective,
                    best_node,
                    initialize_queue,
                    queue_strategy,
                    convergence_checker,
                    node_limit,
                    time_limit,
                    log,
                    log_interval_seconds,
                    log_new_incumbent)
            if not self.is_worker:
                def handler(signum, frame):       #pragma:nocover
                    self._disp.termination_condition = \
                        TerminationCondition.interrupted
                    self._disp.log_warning(
                        "Solve interrupted by user. "
                        "Waiting for current worker "
                        "jobs to complete before "
                        "terminating the solve.")
                with MPI_InterruptHandler(
                        handler,
                        disable=disable_signal_handlers):
                    tmp = self._disp.serve()
            else:
                def handler(signum, frame):       #pragma:nocover
                    if self.is_dispatcher:
                        self._disp.termination_condition = \
                            TerminationCondition.interrupted
                        self._disp.log_warning(
                            "Solve interrupted by user. "
                            "Waiting for current worker "
                            "jobs to complete before "
                            "terminating the solve.")
                with MPI_InterruptHandler(
                        handler,
                        disable=disable_signal_handlers):
                    tmp = self._solve(problem,
                                      best_objective,
                                      best_node,
                                      disable_objective_call,
                                      convergence_checker)
            (results.objective,
             results.best_node,
             results.bound,
             results.termination_condition,
             self._global_solve_info) = tmp
            results.nodes = self._global_solve_info.explored_nodes_count
            self._fill_results(results, convergence_checker)
        except:                                        #pragma:nocover
            sys.stderr.write("Exception caught: "+str(sys.exc_info()[1])+"\n")
            sys.stderr.write("Attempting to shut down, but this may hang.\n")
            sys.stderr.flush()
            raise
        finally:
            problem.load_state(orig)
        self._wall_time = self._time() - self._solve_start
        results.wall_time = self._wall_time

        assert results.solution_status in SolutionStatus,\
            str(results)
        assert results.termination_condition in TerminationCondition,\
            str(results)

        # convert to simple string types
        results.solution_status = results.solution_status.value
        results.termination_condition = results.termination_condition.value

        problem.notify_solve_finished(self.comm,
                                      self.worker_comm,
                                      results)
        if self.is_dispatcher and \
           (log is not None) and \
           (not log.disabled):
            self._disp.log_info("")
            if results.solution_status in ("feasible", "optimal"):
                agap = convergence_checker.compute_absolute_gap(
                    results.bound,
                    results.objective)
                rgap = convergence_checker.compute_relative_gap(
                    results.bound,
                    results.objective)
                if results.solution_status == "feasible":
                    self._disp.log_info("Feasible solution found")
                else:
                    if (convergence_checker.absolute_gap is not None) and \
                       agap <= convergence_checker.absolute_gap:
                        self._disp.log_info("Absolute optimality tolerance met")
                    if (convergence_checker.relative_gap is not None) and \
                       rgap <= convergence_checker.relative_gap:
                        self._disp.log_info("Relative optimality tolerance met")
                    assert results.solution_status == "optimal"
                    self._disp.log_info("Optimal solution found!")
            elif results.solution_status == "infeasible":
                self._disp.log_info("Problem is infeasible")
            elif results.solution_status == "unbounded":
                self._disp.log_info("Problem is unbounded")
            elif results.solution_status == "invalid":      #pragma:nocover
                self._disp.log_info("Problem is invalid")
            else:
                assert results.solution_status == "unknown"
                self._disp.log_info("Status unknown")
            self._disp.log_info("")
            self._disp.log_info(str(results))

        return results

def _nonzero_avg(items, div=None):
    """Returns the average of a list of items, excluding
    zeros. The optional div argument can be set to a list of
    values to divide each item by when computing the
    average."""
    assert (div is None) or \
        (len(items) == len(div))
    s = 0.0
    c = 0
    for i, val in enumerate(items):
        assert val >= 0
        if val != 0:
            div_i = 1 if (div is None) else div[i]
            s += val/float(div_i)
            c += 1
    if c == 0:
        return 0
    return s/float(c)

def summarize_worker_statistics(stats, stream=sys.stdout):
    """Writes a summary of workers statistics to an
    output stream.

    Parameters
    ----------
    stats : dict
        A dictionary of worker statistics returned from
        a call to :func:`collect_worker_statistics`.
    stream : file-like object, or string, optional
        A file-like object or a filename where results
        should be written to. (default: ``sys.stdout``)
    """
    assert all(len(stats[key]) == len(stats['wall_time'])
               for key in stats)
    wall_time = stats['wall_time']
    queue_time = stats['queue_time']
    queue_count = stats['queue_call_count']
    objective_time = stats['objective_time']
    objective_count = stats['objective_call_count']
    bound_time = stats['bound_time']
    bound_count = stats['bound_call_count']
    branch_time = stats['branch_time']
    branch_count = stats['branch_call_count']
    load_state_time = stats['load_state_time']
    load_state_count = stats['load_state_call_count']
    explored_nodes_count = stats['explored_nodes_count']
    work_time = [wt-qt for wt,qt in zip(wall_time,queue_time)]
    sum_enc = sum(explored_nodes_count)
    with as_stream(stream) as stream:
        stream.write("Number of Workers:   %6d\n"
                     % (len(wall_time)))
        if sum_enc == 0:
            stream.write("Load Imbalance:     %6.2f%%\n"
                         % (0.0))
        else:
            max_enc = max(explored_nodes_count)
            min_enc = min(explored_nodes_count)
            avg_enc = sum_enc/float(len(explored_nodes_count))
            stream.write("Load Imbalance:     %6.2f%%\n"
                         % ((max_enc-min_enc)/avg_enc*100.0))
            stream.write(" - min: %d\n" % (min_enc))
            stream.write(" - max: %d\n" % (max_enc))
        stream.write("Average Worker Timing:\n")
        queue_count_str = "%d" % sum(queue_count)
        tmp = "%"+str(len(queue_count_str))+"d"
        bound_count_str = tmp % sum(bound_count)
        objective_count_str = tmp % sum(objective_count)
        branch_count_str = tmp % sum(branch_count)
        load_state_count_str = tmp % sum(load_state_count)
        stream.write(" - queue:     %6.2f%% [avg time: %8s, count: %s]\n"
                     % (_nonzero_avg(queue_time,
                                     div=wall_time)*100.0,
                        time_format(_nonzero_avg(queue_time,
                                                 div=queue_count),
                                    align_unit=True),
                        queue_count_str))
        stream.write(" - load_state:%6.2f%% [avg time: %8s, count: %s]\n"
                     % (_nonzero_avg(load_state_time,
                                     div=wall_time)*100.0,
                        time_format(_nonzero_avg(load_state_time,
                                                 div=load_state_count),
                                    align_unit=True),
                        load_state_count_str))
        stream.write(" - bound:     %6.2f%% [avg time: %8s, count: %s]\n"
                     % (_nonzero_avg(bound_time,
                                     div=wall_time)*100.0,
                        time_format(_nonzero_avg(bound_time,
                                                 div=bound_count),
                                    align_unit=True),
                        bound_count_str))
        stream.write(" - objective: %6.2f%% [avg time: %8s, count: %s]\n"
                     % (_nonzero_avg(objective_time,
                                     div=wall_time)*100.0,
                        time_format(_nonzero_avg(objective_time,
                                                 div=objective_count),
                                    align_unit=True),
                        objective_count_str))
        stream.write(" - branch:    %6.2f%% [avg time: %8s, count: %s]\n"
                     % (_nonzero_avg(branch_time,
                                     div=wall_time)*100.0,
                        time_format(_nonzero_avg(branch_time,
                                                 div=branch_count),
                                    align_unit=True),
                        branch_count_str))
        other_time = [wt-ot-bt-brt-lst if qc != 0 else 0
                      for wt,ot,bt,brt,lst,qc in
                      zip(work_time,
                          objective_time,
                          bound_time,
                          branch_time,
                          load_state_time,
                          queue_count)]
        stream.write(" - other:     %6.2f%% [avg time: %8s, count: %s]\n"
                     % (_nonzero_avg(other_time,
                                     div=wall_time)*100.0,
                        time_format(_nonzero_avg(other_time,
                                                 div=queue_count),
                                    align_unit=True),
                        queue_count_str))

def solve(problem,
          comm=_notset,
          dispatcher_rank=0,
          log_filename=None,
          results_filename=None,
          **kwds):
    """Solves a branch-and-bound problem and returns the
    solution.

    Note
    ----
    This function also collects and summarizes runtime
    workload statistics, which may introduce additional
    overhead. This overhead can be avoided by directly
    instantiating a :class:`Solver` object and
    calling the :func:`Solver.solve` method.

    Parameters
    ----------
    problem : :class:`pybnb.Problem <pybnb.problem.Problem>`
        An object that defines a branch-and-bound problem
    comm : ``mpi4py.MPI.Comm``, optional
        The MPI communicator to use. If unset, the
        mpi4py.MPI.COMM_WORLD communicator will be
        used. Setting this keyword to None will disable the
        use of MPI and avoid an attempted import of
        mpi4py.MPI (which avoids triggering a call to
        `MPI_Init()`).
    dispatcher_rank : int, optional
        The process with this rank will be designated the
        dispatcher process. If MPI functionality is disabled
        (by setting comm=None, or when comm.size==1), this
        keyword must be left at 0. (default: 0)
    log_filename : string, optional
        A filename where solver output should be sent in
        addition to console. This keyword will be ignored if
        the `log` keyword is set. (default: None)
    results_filename : string, optional
        Saves the solver results into a YAML-formatted file
        with the given name. (default: None)
    **kwds
        Additional keywords to be passed to
        :func:`Solver.solve`. See that method for additional
        keyword documentation.

    Returns
    -------
    results : :class:`SolverResults <pybnb.solver_results.SolverResults>`
        An object storing information about the solve.
    """

    opt = Solver(comm=comm,
                 dispatcher_rank=dispatcher_rank)

    if (opt.is_dispatcher) and \
       ("log" not in kwds) and \
       (log_filename is not None):
        kwds["log"] = get_simple_logger(
            filename=log_filename)

    results = opt.solve(problem, **kwds)

    stats = opt.collect_worker_statistics()
    if opt.is_dispatcher:
        tmp = six.StringIO()
        summarize_worker_statistics(stats, stream=tmp)
        opt._disp.log_info(tmp.getvalue())

    if opt.is_dispatcher and (results_filename is not None):
        results.write(results_filename)

    return results

_solve_defaults = get_default_args(Solver.solve)



#
# This example minimizes the 2d Rosenbrock function. It uses
# a spatial decomposition to refine convex relaxations to
# the problem to obtain a lower bound, and uses a locally
# optimal solver to obtain an upper bound. Note that the 2d
# Rosenbrock function is already convex. This problem is
# simply meant to show how these techniques can be applied.
#
# Pyomo version 5.4.3 or later is required to run this
# example.
#

import pybnb

from pybnb.pyomo.problem import PyomoProblem

import pyomo.kernel as pmo

class McCormickEnvelope(pmo.constraint_tuple):
    """A class that stores constraints defining
    the convex envelope for the function: z = xy
    """
    def __init__(self, x, y, z):
        assert x.has_lb() and x.has_ub()
        assert y.has_lb() and y.has_ub()
        self.x = x
        self.y = y
        self.z = z
        super(McCormickEnvelope, self).__init__(
            (pmo.constraint(),
             pmo.constraint(),
             pmo.constraint(),
             pmo.constraint()))
        self.update_constraints()

    def update_constraints(self):
        """Rebuild constraints using current domain variable
        bounds"""
        x, y, z = self.x, self.y, self.z
        assert x.has_lb() and x.has_ub()
        assert y.has_lb() and y.has_ub()
        self[0].body = z - y.lb*x - x.lb*y
        self[0].lb = -x.lb*y.lb
        self[1].body = z - y.ub*x - x.ub*y
        self[1].lb = -x.ub*y.ub
        self[2].body = z - y.ub*x - x.lb*y
        self[2].ub = -x.lb*y.ub
        self[3].body = z - y.lb*x - x.ub*y
        self[3].ub = -x.ub*y.lb

class SquaredEnvelope(pmo.constraint_tuple):
    """A class that stores constraints defining
    the convex envelope for the function: z = x^2"""

    def __init__(self, x, z):
        assert x.has_lb() and x.has_ub()
        self.x = x
        self.z = z
        super(SquaredEnvelope, self).__init__(
            (pmo.constraint(),
             pmo.constraint()))
        self[0].body = self.z - self.x**2
        self[0].lb = 0
        self.update_constraints()

    def update_constraints(self):
        """Rebuild constraints using current domain variable
        bounds"""
        x, z = self.x, self.z
        assert x.has_lb() and x.has_ub()
        self[1].body = z - (x.lb + x.ub)*x
        self[1].ub = -x.lb*x.ub

    def derived_output_bounds(self):
        x = self.x
        assert x.has_lb() and x.has_ub()
        vals = (x.lb**2, x.ub**2)
        if (x.lb <= 0) and (x.ub >= 0):
            lb = 0.0
        else:
            lb = min(vals)
        ub = max(vals)
        return lb, ub

class Rosenbrock2D(PyomoProblem):

    def __init__(self, xL, xU, yL, yU):
        assert xL <= xU
        assert yL <= yU
        self._model = pmo.block()
        x = self._model.x = pmo.variable(lb=xL, ub=xU)
        y = self._model.y = pmo.variable(lb=yL, ub=yU)
        x2 = self._model.x2 = pmo.variable(
            lb=-pybnb.inf,
            ub=pybnb.inf)
        x2y = self._model.x2y = pmo.variable(
            lb=-pybnb.inf,
            ub=pybnb.inf)
        self._model.x2_c = SquaredEnvelope(x, x2)
        # Temporarily provide bounds to the x2 variable so
        # they can be used to build the McCormick
        # constraints for x2y. After that, they are no
        # longer needed as they are enforced by the
        # McCormickEnvelope constraints.
        x2.bounds = self._model.x2_c.derived_output_bounds()
        self._model.x2y_c = McCormickEnvelope(x2, y, x2y)
        x2.bounds = (-pybnb.inf,
                     pybnb.inf)

        # original objective
        self._model.f = pmo.expression(
            (x**4 - 2*(x**2)*y + \
             0.5*(x**2) - x + \
             (y**2) + 0.5))
        # convex relaxation
        self._model.f_convex = pmo.expression(
            (x**4 - 2*x2y + \
             0.5*(x**2) - x + \
             (y**2) + 0.5))
        self._model.objective = pmo.objective(sense=pmo.minimize)
        self._ipopt = pmo.SolverFactory("ipopt")
        self._ipopt.options['tol'] = 1e-9
        self._last_bound_was_feasible = False

        # make sure the PyomoProblem initializer is called
        # after the model is built
        super(Rosenbrock2D, self).__init__()

    def check_feasible(self):
        """Check if the currently loaded solution
        is feasible for the true model."""
        # assumes the relaxed model was just solved
        assert not self._model.x2y.stale
        x = self._model.x
        y = self._model.y
        resid = 0.0
        # x.lb <= x <= x.ub
        resid += min(0,x.slack)**2
        # y.lb <= y <= y.ub
        resid += min(0,y.slack)**2
        # x2y <= y(x^2)
        resid += min(0, (x.value*x.value*y.value) - \
                        self._model.x2y.value)**2
        resid = resid**(0.5)
        if resid < 1e-5:
            return True
        else:
            return False

    def setup_model_for_objective(self):
        self._model.objective.expr = self._model.f
        self._model.x2_c.deactivate()
        self._model.x2y_c.deactivate()

    def setup_model_for_bound(self):
        self._model.objective.expr = self._model.f_convex
        self._model.x2_c.activate()
        self._model.x2y_c.activate()

    def rebuild_convex_envelopes(self):
        self._model.x2_c.update_constraints()
        # Temporarily provide bounds to the x2 variable so
        # they can be used to build the McCormick
        # constraints for x2y. After that, they are no
        # longer needed as they are enforced by the
        # SquaredEnvelope constraints.
        self._model.x2.bounds = \
            self._model.x2_c.derived_output_bounds()
        self._model.x2y_c.update_constraints()
        self._model.x2.bounds = (-pybnb.inf,
                                 pybnb.inf)

    #
    # Implement PyomoProblem abstract methods
    #

    @property
    def pyomo_model(self):
        return self._model

    @property
    def pyomo_model_objective(self):
        return self._model.objective

    #
    # Implement Problem abstract methods
    #

    #def sense(self): # implemented by PyomoProblem base class

    def objective(self):
        self.setup_model_for_objective()
        results = self._ipopt.solve(self._model, load_solutions=False)
        if (str(results.solver.status) == "ok") and \
           (str(results.solver.termination_condition) == "optimal"):
            self._model.load_solution(results.solution(0))
            assert self._model.x2y.stale
            return round(self._model.objective(), 7)
        else:                                     #pragma:nocover
            assert str(results.solver.status) == "warning"
            if str(results.solver.termination_condition) == "unbounded":
                return self.unbounded_objective()
            else:
                assert str(results.solver.termination_condition) == "infeasible"
                return self.infeasible_objective()

    def bound(self):
        self.setup_model_for_bound()
        self._last_bound_was_feasible = False
        results = self._ipopt.solve(self._model, load_solutions=False)
        if (str(results.solver.status) == "ok") and \
           (str(results.solver.termination_condition) == "optimal"):
            self._model.load_solution(results.solution(0))
            self._last_bound_was_feasible = self.check_feasible()
            return round(self._model.objective(), 7)
        else:                                     #pragma:nocover
            assert str(results.solver.status) == "warning"
            if str(results.solver.termination_condition) == "unbounded":
                return self.unbounded_objective()
            else:
                assert str(results.solver.termination_condition) == "infeasible"
                return self.infeasible_objective()

    def save_state(self, node):
        node.state = (self._model.x.bounds,
                      self._model.y.bounds)

    def load_state(self, node):
        assert len(node.state) == 2
        self._model.x.bounds = node.state[0]
        self._model.y.bounds = node.state[1]
        self.rebuild_convex_envelopes()
        self._last_bound_was_feasible = False

    def branch(self):
        if self._last_bound_was_feasible:
            return ()
        xL, xU = self._model.x.bounds
        yL, yU = self._model.y.bounds
        xdist = float(xU-xL)
        ydist = float(yU-yL)
        branch_var = None
        if xdist > ydist:
            branch_var = self._model.x
            L = xL
            U = xU
        else:
            branch_var = self._model.y
            L = yL
            U = yU
        # branch
        mid = 0.5*(L + U)
        left = pybnb.Node()
        branch_var.bounds = (L, mid)
        self.save_state(left)
        right = pybnb.Node()
        branch_var.bounds = (mid, U)
        self.save_state(right)
        # reset the variable bounds
        branch_var.bounds = (L, U)
        return [left, right]

if __name__ == "__main__":
    import pybnb.misc

    problem = Rosenbrock2D(xL=-25, xU=25,
                           yL=-25 ,yU=25)
    pybnb.misc.create_command_line_solver(problem)



import math

from pybnb.common import (minimize,
                          maximize,
                          inf,
                          nan,
                          QueueStrategy,
                          _queue_strategy_to_int,
                          _int_to_queue_strategy,
                          SolutionStatus,
                          _solution_status_to_int,
                          _int_to_solution_status,
                          TerminationCondition,
                          _termination_condition_to_int,
                          _int_to_termination_condition)

class Test(object):

    def test_minimize(self):
        assert minimize == 1

    def test_maximize(self):
        assert maximize == -1

    def test_inf(self):
        assert math.isinf(inf)
        assert math.isinf(-inf)

    def test_nan(self):
        assert math.isnan(nan)
        assert math.isnan(-nan)

    def test_QueueStrategy(self):
        assert QueueStrategy.bound == "bound"
        assert QueueStrategy.objective == "objective"
        assert QueueStrategy.breadth == "breadth"
        assert QueueStrategy.depth == "depth"
        assert QueueStrategy.local_gap == "local_gap"
        assert QueueStrategy.fifo == "fifo"
        assert QueueStrategy.lifo == "lifo"
        assert QueueStrategy.random == "random"
        assert QueueStrategy.custom == "custom"
        assert len(QueueStrategy) == 9
        assert len(_queue_strategy_to_int) == 9
        assert len(_int_to_queue_strategy) == 9
        for val in QueueStrategy:
            assert _int_to_queue_strategy[
                _queue_strategy_to_int[val]] is val

    def test_SolutionStatus(self):
        assert SolutionStatus.optimal == "optimal"
        assert SolutionStatus.feasible == "feasible"
        assert SolutionStatus.infeasible == "infeasible"
        assert SolutionStatus.unbounded == "unbounded"
        assert SolutionStatus.invalid == "invalid"
        assert SolutionStatus.unknown == "unknown"
        assert len(SolutionStatus) == 6
        assert len(_solution_status_to_int) == 6
        assert len(_int_to_solution_status) == 6
        for i, val in enumerate(SolutionStatus):
            assert _int_to_solution_status[
                _solution_status_to_int[val]] is val

    def test_TerminationCondition(self):
        assert TerminationCondition.optimality == "optimality"
        assert TerminationCondition.objective_limit == "objective_limit"
        assert TerminationCondition.node_limit == "node_limit"
        assert TerminationCondition.time_limit == "time_limit"
        assert TerminationCondition.no_nodes == "no_nodes"
        assert TerminationCondition.interrupted == "interrupted"
        assert len(TerminationCondition) == 6
        assert len(_termination_condition_to_int) == 6
        assert len(_int_to_termination_condition) == 6
        for i, val in enumerate(TerminationCondition):
            assert _int_to_termination_condition[
                _termination_condition_to_int[val]] is val



#
# This example solves a simple one dimensional problem that
# has an expensive objective function. A bound is computed
# using a specified lipschitz constant (assumed to be
# valid).
#
# The package numba is not required, but it will make
# this example run faster.
#

import math

import pybnb

try:
    from numba import jit
except ImportError:
    jit = lambda x: x

class Lipschitz1D(pybnb.Problem):

    _LC = 12.0

    def __init__(self, xL, xU):
        assert xL <= xU
        self._xL = xL
        self._xM = 0.5*(xL+xU)
        self._xU = xU
        assert self._xL <= self._xM <= self._xU
        self._fL_cached = None
        self._fM_cached = None
        self._fU_cached = None

    @jit
    def _f(self, x):
        ans = 0
        for i in range(1000,0,-1):
            temp = 0
            for j in range(i,0,-1):
                temp += (x + j)**(-3.1)
            ans += math.sin(x + temp) / (1.2**i)
        return ans

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return pybnb.maximize

    def objective(self):
        if self._fM_cached is None:
            self._fM_cached = self._f(self._xM)
        return self._fM_cached

    def bound(self):
        if self._fL_cached is None:
            self._fL_cached = self._f(self._xL)
        if self._fU_cached is None:
            self._fU_cached = self._f(self._xU)
        return 0.5*self._fL_cached + \
               0.5*self._fU_cached + \
               0.5*self._LC*(self._xU-self._xL)

    def save_state(self, node):
        node.state = (self._xL,
                      self._xU,
                      self._fL_cached,
                      self._fU_cached)

    def load_state(self, node):
        (self._xL,
         self._xU,
         self._fL_cached,
         self._fU_cached) = node.state
        self._xM = 0.5*(self._xL+self._xU)
        self._fM_cached = None
        assert self._xL <= self._xM <= self._xU

    def branch(self):
        child = pybnb.Node()
        child.state = (self._xL,
                       self._xM,
                       self._fL_cached,
                       self._fM_cached)
        yield child
        child = pybnb.Node()
        child.state = (self._xM,
                       self._xU,
                       self._fM_cached,
                       self._fU_cached)
        yield child

if __name__ == "__main__":
    import pybnb.misc

    problem = Lipschitz1D(0,10)
    pybnb.misc.create_command_line_solver(problem)



"""
A proxy interface to the central dispatcher that is used by
branch-and-bound workers.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import array
import collections
import marshal

from pybnb.configuration import config
from pybnb.common import _int_to_termination_condition
from pybnb.node import _SerializedNode
from pybnb.problem import _SolveInfo
from pybnb.mpi_utils import (send_nothing,
                             recv_data)

try:
    import mpi4py
except ImportError:                               #pragma:nocover
    pass

import six

_ProcessType = collections.namedtuple(
    "_ProcessType",
    ["worker",
     "dispatcher"])
ProcessType = _ProcessType(
    worker     = 0,
    dispatcher = 1)
"""A namespace of typecodes that are used to categorize
processes during dispatcher startup."""

_DispatcherAction = collections.namedtuple(
    "_DispatcherAction",
    ["update",
     "finalize",
     "log_info",
     "log_warning",
     "log_debug",
     "log_error",
     "log_critical",
     "stop_listen"])

DispatcherAction = _DispatcherAction(
    update                    = 111,
    finalize                  = 211,
    log_info                  = 311,
    log_warning               = 411,
    log_debug                 = 511,
    log_error                 = 611,
    log_critical              = 711,
    stop_listen               = 811)
"""A namespace of typecodes that are used to categorize
messages received by the dispatcher from workers."""

_DispatcherResponse = collections.namedtuple(
    "_DispatcherResponse",
    ["work",
     "nowork"])
DispatcherResponse = _DispatcherResponse(
    work             = 1111,
    nowork           = 2111)
"""A namespace of typecodes that are used to categorize
responses received by workers from the dispatcher."""

class DispatcherProxy(object):
    """A proxy class for interacting with the central
    dispatcher via message passing."""

    @staticmethod
    def _init(comm, ptype):
        """Broadcasts the dispatcher rank to everyone and
        sets up a worker communicator that excludes the
        single dispatcher."""
        import mpi4py.MPI
        assert mpi4py.MPI.Is_initialized()
        assert len(ProcessType) == 2
        assert ProcessType.dispatcher == 1
        assert ProcessType.worker == 0
        assert ptype in ProcessType
        ptype_, dispatcher_rank = comm.allreduce(
            sendobj=(ptype, comm.rank),
            op=mpi4py.MPI.MAXLOC)
        assert ptype_ == ProcessType.dispatcher
        color = None
        if ptype == ProcessType.dispatcher:
            assert dispatcher_rank == comm.rank
            color = 1
        else:
            assert dispatcher_rank != comm.rank
            color = 0
        assert color is not None
        worker_comm = comm.Split(color)
        if color == 1:
            worker_comm.Free()
            return dispatcher_rank
        else:
            return dispatcher_rank, worker_comm

    def __init__(self, comm):
        import mpi4py.MPI
        assert mpi4py.MPI.Is_initialized()
        self.comm = comm
        self.worker_comm = None
        self._status = mpi4py.MPI.Status()
        (self.dispatcher_rank,
         self.worker_comm) = self._init(comm, ProcessType.worker)

    def __del__(self):
        if self.worker_comm is not None:
            self.worker_comm.Free()
            self.worker_comm = None

    def update(self,
               best_objective,
               best_node,
               previous_bound,
               solve_info,
               node_list):
        """A proxy to :func:`pybnb.dispatcher.Dispatcher.update`."""
        if best_node is not None:
            best_node = _SerializedNode.to_slots(best_node)
        node_list = [_SerializedNode.to_slots(node_)
                     for node_ in node_list]
        data = marshal.dumps((best_objective,
                              best_node,
                              previous_bound,
                              solve_info.data,
                              node_list),
                             config.MARSHAL_PROTOCOL_VERSION)
        self.comm.Send([data,mpi4py.MPI.BYTE],
                       self.dispatcher_rank,
                       tag=DispatcherAction.update)
        self.comm.Probe(status=self._status)
        assert not self._status.Get_error()
        tag = self._status.Get_tag()
        recv_size = self._status.Get_count(mpi4py.MPI.BYTE)
        data = bytearray(recv_size)
        recv_data(self.comm,
                  self._status,
                  datatype=mpi4py.MPI.BYTE,
                  out=data)
        if tag == DispatcherResponse.nowork:
            if six.PY2:
                data_ = str(data)
            else:
                data_ = data
            (best_objective,
             best_node_slots,
             global_bound,
             termination_condition_int,
             solve_info_data) = marshal.loads(data_)
            best_node = None
            if best_node_slots is not None:
                best_node = _SerializedNode.restore_node(
                    best_node_slots)
            solve_info = _SolveInfo()
            solve_info.data = array.array('d',solve_info_data)
            return (True,
                    best_objective,
                    best_node,
                    (global_bound,
                     _int_to_termination_condition[
                         termination_condition_int],
                     solve_info))
        else:
            assert tag == DispatcherResponse.work
            if six.PY2:
                data_ = str(data)
            else:
                data_ = data
            (best_objective,
             best_node_slots,
             node_slots) = marshal.loads(data_)
            best_node = None
            if best_node_slots is not None:
                best_node = _SerializedNode.restore_node(
                    best_node_slots)
            node = _SerializedNode.restore_node(node_slots)
            return False, best_objective, best_node, node

    def log_info(self, msg):
        """A proxy to :func:`pybnb.dispatcher.Dispatcher.log_info`."""
        self.comm.Ssend([msg.encode("utf8"),mpi4py.MPI.CHAR],
                        self.dispatcher_rank,
                        tag=DispatcherAction.log_info)

    def log_warning(self, msg):
        """A proxy to :func:`pybnb.dispatcher.Dispatcher.log_warning`."""
        self.comm.Ssend([msg.encode("utf8"),mpi4py.MPI.CHAR],
                        self.dispatcher_rank,
                        tag=DispatcherAction.log_warning)

    def log_debug(self, msg):
        """A proxy to :func:`pybnb.dispatcher.Dispatcher.log_debug`."""
        self.comm.Ssend([msg.encode("utf8"),mpi4py.MPI.CHAR],
                        self.dispatcher_rank,
                        tag=DispatcherAction.log_debug)

    def log_error(self, msg):
        """A proxy to :func:`pybnb.dispatcher.Dispatcher.log_error`."""
        self.comm.Ssend([msg.encode("utf8"),mpi4py.MPI.CHAR],
                        self.dispatcher_rank,
                        tag=DispatcherAction.log_error)

    def log_critical(self, msg):
        """A proxy to :func:`pybnb.dispatcher.Dispatcher.log_critical`."""
        self.comm.Ssend([msg.encode("utf8"),mpi4py.MPI.CHAR],
                        self.dispatcher_rank,
                        tag=DispatcherAction.log_critical)

    def stop_listen(self):
        """Tell the dispatcher to abruptly stop the listen loop."""
        assert self.worker_comm.rank == 0
        send_nothing(self.comm,
                     self.dispatcher_rank,
                     tag=DispatcherAction.stop_listen)
